import io
import os
import tempfile
from pathlib import Path
from typing import IO, Any, Dict, Iterator, Union
from unittest.mock import patch
from uuid import uuid4

import httpx
import pandas as pd
import pyarrow.dataset as ds
import pytest

from avatars.api import Datasets, PandasIntegration
from avatars.conftest import RequestHandle, api_client_factory
from avatars.models import Dataset, FileType

TEST_MAX_BYTES_PER_FILE = 1 * 1024  # 1 KB


@pytest.fixture(scope="session")
def dataset_json() -> dict[str, Any]:
    return {
        "id": "5e27da4a-a5c5-458b-988e-ef2c66f545d6",
        "hash": "aeedab1ee7a1043753c9ab768594bc8420d7b85491d0be9421edc3813c237f4c",
        "name": "upload",
        "download_url": "http://localhost:8000/datasets/5e27da4a-a5c5-458b-988e-ef2c66f545d6/download",  # noqa: E501
        "nb_lines": 1,
        "nb_dimensions": 2,
        "filetype": "csv",
        "columns": [],
    }


@pytest.fixture(scope="session")
def csv_content() -> bytes:
    return b"a,b\n1,2"


@pytest.fixture(scope="session")
def parquet_content(csv_content: bytes) -> bytes:
    result: bytes = pd.read_csv(io.BytesIO(csv_content)).to_parquet()
    return result


class TestCustomCreateDatasetMethod:
    @pytest.fixture(scope="session")
    def create_dataset_response(self, dataset_json: dict[str, Any]) -> RequestHandle:
        def handler(request: httpx.Request) -> httpx.Response:
            return httpx.Response(200, json=dataset_json)

        return handler

    @pytest.fixture(scope="session")
    def large_csv(self) -> bytes:
        return b"a,b\n" + b"1,2\n" * TEST_MAX_BYTES_PER_FILE

    @pytest.mark.parametrize("content_fixture_name", ["csv_content", "parquet_content"])
    def test_create_dataset_from_stream(
        self,
        create_dataset_response: RequestHandle,
        content_fixture_name: str,
        request: Any,
    ) -> None:
        """Verify the call is deprecated, but it succeeds."""
        # TODO: remove this test when create_dataset_from_stream is removed
        content = request.getfixturevalue(content_fixture_name)
        client = api_client_factory(create_dataset_response)
        with pytest.deprecated_call(match="create_dataset_from_stream is deprecated"):
            dataset = Datasets(client).create_dataset_from_stream(
                request=io.BytesIO(content)
            )

        assert dataset.id

    @pytest.mark.parametrize("content_fixture_name", ["csv_content", "parquet_content"])
    def test_create_dataset_request_argument_is_deprecated(
        self,
        create_dataset_response: RequestHandle,
        content_fixture_name: str,
        request: Any,
    ) -> None:
        """Verify the parameter is deprecated, but it succeeds."""
        content = request.getfixturevalue(content_fixture_name)

        client = api_client_factory(create_dataset_response)
        with pytest.deprecated_call(match="request is deprecated"):
            dataset = Datasets(client).create_dataset(request=io.BytesIO(content))

        assert dataset.id

    @pytest.mark.filterwarnings(
        "ignore:request is deprecated:DeprecationWarning"
    )  # TODO: Remove
    def test_create_dataset_both_request_and_source_raises(self) -> None:
        client = api_client_factory()
        with pytest.raises(ValueError, match="You cannot pass both request and source"):
            Datasets(client).create_dataset(request=io.BytesIO(), source=io.BytesIO())

    def test_create_dataset_neither_request_nor_source_raises(self) -> None:
        client = api_client_factory()
        with pytest.raises(ValueError, match="You need to pass in a source"):
            Datasets(client).create_dataset()

    def test_create_dataset_with_unknown_source_type(self) -> None:
        client = api_client_factory()
        with pytest.raises(TypeError, match="Unsupported dataset source"):
            Datasets(client).create_dataset(source=1)

    @pytest.mark.parametrize("content_fixture_name", ["csv_content", "parquet_content"])
    def test_create_dataset_using_source_argument_with_filename(
        self,
        create_dataset_response: RequestHandle,
        content_fixture_name: str,
        request: Any,
    ) -> None:
        # Arrange
        content = request.getfixturevalue(content_fixture_name)
        client = api_client_factory(create_dataset_response)
        with tempfile.NamedTemporaryFile() as tmp_file:
            tmp_file.write(content)
            tmp_file.flush()
            result = Datasets(client).create_dataset(source=tmp_file.name)

        assert result.id

    @patch("avatars.api.MAX_BYTES_PER_FILE", TEST_MAX_BYTES_PER_FILE)
    @pytest.mark.parametrize(
        "filetype, buffer",
        [
            (FileType.csv, io.BytesIO),
            (FileType.csv, io.StringIO),  # TODO: Remove when request is deprecated
            (FileType.parquet, io.BytesIO),
        ],
    )
    def test_create_dataset_with_too_large_buffer_ok(
        self,
        create_dataset_response: RequestHandle,
        filetype: FileType,
        buffer: Any,
        large_csv: bytes,
    ) -> None:
        client = api_client_factory(create_dataset_response)

        filled_buffer: Union[io.BytesIO, io.StringIO]
        if filetype == FileType.csv:
            content = large_csv if buffer == io.BytesIO else large_csv.decode()
            filled_buffer = buffer(content)
        else:
            filled_buffer = buffer(pd.read_csv(io.BytesIO(large_csv)).to_parquet())

        filled_buffer.seek(0)
        # TODO: Remove the type ignore when request is deprecated
        # error: Argument "source" to "create_dataset" of "Datasets"
        #        has incompatible type "Union[BytesIO, StringIO]"; expected
        #        "Union[str, list[str], FileLikeInterface[bytes], None]"  [arg-type]
        res = Datasets(client).create_dataset(source=filled_buffer)
        assert res.id

    @patch("avatars.api.MAX_BYTES_PER_FILE", TEST_MAX_BYTES_PER_FILE)
    @pytest.mark.parametrize(
        "filetype, mode",
        [
            (
                FileType.csv,
                "rb",
            ),
            (FileType.parquet, "rb"),
        ],
        ids=str,
    )
    def test_create_dataset_with_too_large_file_ok(
        self,
        create_dataset_response: RequestHandle,
        filetype: FileType,
        mode: str,
        large_csv: bytes,
    ) -> None:
        client = api_client_factory(create_dataset_response)

        buffer = io.BytesIO(large_csv)
        buffer.seek(0)
        df = pd.read_csv(buffer)
        with tempfile.NamedTemporaryFile() as tmp_file:
            if filetype == FileType.parquet:
                df.to_parquet(tmp_file.name)
            else:
                df.to_csv(tmp_file.name)

            file_ = open(tmp_file.name, mode)
            res = Datasets(client).create_dataset(source=file_)
            assert res.id

    @pytest.mark.filterwarnings(
        "ignore:You are trying to upload a text file:DeprecationWarning"
    )
    @pytest.mark.parametrize("content_fixture_name", ["csv_content", "parquet_content"])
    @pytest.mark.parametrize("source", [io.BytesIO, io.StringIO])
    def test_create_dataset_using_source_argument_with_buffer(
        self,
        source: type[io.IOBase],
        content_fixture_name: str,
        create_dataset_response: RequestHandle,
        request: Any,
    ) -> None:
        if source == io.StringIO and content_fixture_name == "parquet_content":
            # We're skipping it here, because it makes it easier to handle a single test
            # with parametrized fixtures, rather than having to create a separate test
            # for each combination of source and content type.
            pytest.skip(
                "Parquet files are not UTF-8 encoded and cannot be read as a string."
            )

        client = api_client_factory(create_dataset_response)
        content = request.getfixturevalue(content_fixture_name)
        if source == io.StringIO:
            buffer = source(content.decode())  # type: ignore[call-arg]
        else:
            buffer = source(content)  # type: ignore[call-arg]

        res = Datasets(client).create_dataset(source=buffer)
        assert res.id

    def test_create_dataset_using_source_argument_with_multiple_files(
        self, create_dataset_response: RequestHandle, parquet_content: bytes
    ) -> None:
        # Arrange
        client = api_client_factory(create_dataset_response)

        # Act
        with tempfile.TemporaryDirectory() as tmpdir:
            filename_1 = Path(tmpdir).joinpath("file1.txt")
            filename_2 = Path(tmpdir).joinpath("file2.txt")
            filename_1.write_bytes(parquet_content)
            filename_2.write_bytes(parquet_content)
            res = Datasets(client).create_dataset(
                source=[str(filename_1), str(filename_2)]
            )
        assert res.id

    @pytest.mark.filterwarnings(
        "ignore:request is deprecated:DeprecationWarning"
    )  # TODO: Remove
    def test_create_dataset_with_deprecated_request_argument_calls_correct_request_method(
        self, create_dataset_response: RequestHandle, dataset_json: dict[str, Any]
    ) -> None:
        client = api_client_factory(create_dataset_response)

        with patch.object(client, "request", return_value=dataset_json) as mock_request:
            # as positional argument
            Datasets(client).create_dataset(io.BytesIO(b"123\n"))
            mock_request.assert_called_once()

            ds_ = mock_request.call_args.kwargs["dataset"]

            assert isinstance(ds_, ds.Dataset)

            mock_request.reset_mock()

            # as keyword argument
            Datasets(client).create_dataset(request=io.BytesIO(b"123\n"))
            mock_request.assert_called_once()

            ds_ = mock_request.call_args.kwargs["dataset"]

            assert isinstance(ds_, ds.Dataset)


@pytest.fixture(scope="session")
def download_dataset_parquet_response(parquet_content: bytes) -> httpx.Response:
    return httpx.Response(
        200,
        content=parquet_content,
        headers={"content-type": "application/octet-stream"},
    )


@pytest.fixture(scope="session")
def download_dataset_csv_response(csv_content: bytes) -> httpx.Response:
    return httpx.Response(
        200,
        content=csv_content,
        headers={"content-type": "text/csv"},
    )


@pytest.fixture(scope="session")
def get_dataset_response(dataset_json: Dict[str, Any]) -> httpx.Response:
    return httpx.Response(200, json=dataset_json)


@pytest.fixture(scope="session")
def download_dataset_response(
    get_dataset_response: httpx.Response,
    download_dataset_parquet_response: httpx.Response,
    download_dataset_csv_response: httpx.Response,
) -> RequestHandle:
    def handler(request: httpx.Request) -> httpx.Response:
        is_download_dataset = "download" in request.url.path
        is_get_dataset = not is_download_dataset and request.url.path.startswith(
            "/datasets"
        )
        if is_download_dataset:
            is_parquet_filetype = b"parquet" in request.url.query
            if is_parquet_filetype:
                return download_dataset_parquet_response
            else:
                return download_dataset_csv_response
        elif is_get_dataset:
            return get_dataset_response
        else:
            raise ValueError("Unexpected request")

    return handler


class TestCustomDownloadDatasetMethod:
    @pytest.mark.parametrize(
        "filetype,expected_output_fixture_name",
        [
            (FileType.csv, "csv_content"),
            (FileType.parquet, "parquet_content"),
        ],
    )
    def test_download_dataset_as_stream(
        self,
        download_dataset_response: RequestHandle,
        filetype: FileType,
        expected_output_fixture_name: str,
        request: Any,
    ) -> None:
        """Verify the call is deprecated, but it succeeds."""
        # TODO: remove this test when download_dataset_as_stream is removed
        client = api_client_factory(download_dataset_response)
        with pytest.deprecated_call(match="download_dataset_as_stream is deprecated"):
            response = Datasets(client).download_dataset_as_stream(
                str(uuid4()), filetype=filetype
            )

        expected = request.getfixturevalue(expected_output_fixture_name)
        assert isinstance(response, io.BytesIO)
        assert response.read() == expected

    @pytest.mark.parametrize(
        "filetype,expected_output_fixture_name",
        [
            (FileType.csv, "csv_content"),
            (None, "csv_content"),  # old behavior
        ],
    )
    def test_download_dataset_when_destination_is_none(
        self,
        download_dataset_response: RequestHandle,
        filetype: FileType,
        expected_output_fixture_name: str,
        request: Any,
    ) -> None:
        """Verify that we raise a DeprecationWarning, but keep the old behavior."""
        client = api_client_factory(download_dataset_response)
        with pytest.warns(
            DeprecationWarning, match="Please specify the destination argument"
        ):
            response = Datasets(client).download_dataset(
                str(uuid4()),
                destination=None,
                from_download_as_stream=False,
                filetype=filetype,
            )

            expected: bytes = request.getfixturevalue(expected_output_fixture_name)
            assert isinstance(response, str)
            assert response == expected.decode()

    @pytest.mark.parametrize(
        "filetype,expected_output_fixture_name",
        [
            (FileType.csv, "csv_content"),
            (FileType.parquet, "parquet_content"),
        ],
    )
    def test_download_dataset_with_valid_destination_as_filename(
        self,
        download_dataset_response: RequestHandle,
        filetype: FileType,
        expected_output_fixture_name: str,
        request: Any,
    ) -> None:
        """Verify that the method works correctly with a valid destination."""
        client = api_client_factory(download_dataset_response)
        with tempfile.NamedTemporaryFile() as tmp:
            Datasets(client).download_dataset(
                str(uuid4()), destination=tmp.name, filetype=filetype
            )

            expected: bytes = request.getfixturevalue(expected_output_fixture_name)
            assert Path(tmp.name).read_bytes() == expected

    @pytest.fixture(scope="function")
    def binary_file_handle(self) -> Iterator[IO[bytes]]:
        new_file = tempfile.NamedTemporaryFile("rb+")
        yield new_file
        new_file.close()

    @pytest.fixture(scope="function")
    def file_handle(self) -> Iterator[IO[str]]:
        new_file = tempfile.NamedTemporaryFile("r+")
        yield new_file
        new_file.close()

    @pytest.mark.parametrize(
        "filetype,expected_output_fixture_name,destination",
        [
            pytest.param(
                FileType.csv, "csv_content", io.BytesIO(), id="io.BytesIO-csv"
            ),
            pytest.param(
                FileType.csv, "csv_content", io.StringIO(), id="io.StringIO-csv"
            ),
            pytest.param(FileType.csv, "csv_content", "binary_file_handle"),
            pytest.param(FileType.csv, "csv_content", "file_handle"),
            pytest.param(
                FileType.parquet,
                "parquet_content",
                io.BytesIO(),
                id="io.BytesIO-parquet",
            ),
            pytest.param(FileType.parquet, "parquet_content", "binary_file_handle"),
        ],
    )
    def test_download_dataset_with_valid_destination_as_buffer(
        self,
        download_dataset_response: RequestHandle,
        destination: Union[str, IO[bytes], IO[str]],
        filetype: FileType,
        expected_output_fixture_name: str,
        request: Any,
    ) -> None:
        """Verify that the method works correctly with a valid destination."""
        client = api_client_factory(download_dataset_response)
        dataset_id = str(uuid4())

        # grab the open file handle fixture
        destination = (
            request.getfixturevalue(destination)
            if isinstance(destination, str)
            else destination
        )

        response = Datasets(client).download_dataset(
            dataset_id,
            destination=destination,
            filetype=filetype,
        )

        assert response is None

        expected: bytes = request.getfixturevalue(expected_output_fixture_name)

        destination.seek(0, os.SEEK_SET)
        actual = destination.read()
        actual = actual if isinstance(actual, bytes) else actual.encode()

        assert actual == expected

    def test_download_dataset_with_invalid_destination_fails(
        self,
        download_dataset_response: RequestHandle,
    ) -> None:
        """Verify that the method raises an error with an invalid destination."""
        client = api_client_factory(download_dataset_response)
        with pytest.raises(
            TypeError, match="Expected destination to be a string or a buffer"
        ):
            Datasets(client).download_dataset(
                str(uuid4()),
                destination=123,  # type: ignore[arg-type]
            )


class TestPandasIntegrationUploadDataframe:
    @pytest.fixture
    def dataframe(self) -> pd.DataFrame:
        return pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})

    @pytest.fixture(scope="session")
    def create_dataset_response(self, dataset_json: dict[str, Any]) -> RequestHandle:
        def handler(request: httpx.Request) -> httpx.Response:
            return httpx.Response(200, json=dataset_json)

        return handler

    @pytest.mark.filterwarnings(
        "ignore:You are trying to upload a text file:DeprecationWarning"
    )
    def test_upload_dataframe(
        self, dataframe: pd.DataFrame, create_dataset_response: RequestHandle
    ) -> None:
        # Arrange
        client = api_client_factory(create_dataset_response)
        # Act
        result = PandasIntegration(client).upload_dataframe(dataframe)
        # Assert
        assert isinstance(result, Dataset)


class TestPandasIntegrationDownloadDataframe:
    @pytest.fixture
    def dataset(self, dataset_json: dict[str, Any]) -> Dataset:
        return Dataset(**dataset_json)

    @pytest.fixture(scope="session")
    def dataframe(self, csv_content: bytes) -> pd.DataFrame:
        return pd.read_csv(io.BytesIO(csv_content))

    @pytest.mark.filterwarnings(
        "ignore:download_dataset_as_stream is deprecated:DeprecationWarning"
    )
    def test_download_dataframe(
        self,
        download_dataset_response: RequestHandle,
        dataset: Dataset,
        dataframe: pd.DataFrame,
    ) -> None:
        """Verify the call is deprecated, but it succeeds."""
        client = api_client_factory(download_dataset_response)
        with pytest.deprecated_call(
            match="The `should_stream` parameter is deprecated"
        ):
            result = PandasIntegration(client).download_dataframe(
                str(dataset.id), should_stream=True
            )

        pd.testing.assert_frame_equal(result, dataframe)



import numpy as np
import pandas as pd
import pytest

from avatars.processors import GeolocationNormalizationProcessor


@pytest.fixture
def df() -> pd.DataFrame:
    df = pd.read_csv("./fixtures/porto_taxi_200.csv")
    return df


def test_preprocess(df: pd.DataFrame) -> None:
    processor = GeolocationNormalizationProcessor(
        latitude_variable='lat',
        longitude_variable='lon',
        n_reference_lat=10,
        n_bins=5)
    processed_df = processor.preprocess(df=df)

    assert len(processed_df) == len(df)


def test_postprocess(df: pd.DataFrame) -> None:
    processor = GeolocationNormalizationProcessor(
        latitude_variable='lat',
        longitude_variable='lon',
        n_reference_lat=10,
        n_bins=5)
    random_points_df = df.sample(frac=0.25, replace=True).reset_index(drop=True)
    random_points_df['lon'] = np.random.uniform(0.0, 1.0, len(random_points_df))

    processor.preprocess(df=df)
    postprocessed_df = processor.postprocess(source=df, dest=random_points_df)

    assert len(postprocessed_df) == len(random_points_df)
    assert min(postprocessed_df['lon']) >= min(df['lon'])
    assert max(postprocessed_df['lon']) <= max(df['lon'])



import inspect
import io
import os
import struct
from contextlib import contextmanager
from dataclasses import dataclass, field
from enum import Enum
from io import BytesIO
from pathlib import Path
from typing import (
    IO,
    Any,
    AsyncGenerator,
    BinaryIO,
    Callable,
    Generator,
    Iterator,
    Optional,
    Union,
)

import clevercsv
import pandas as pd
import pyarrow as pa
import pyarrow.csv as pcsv
import pyarrow.dataset as ds
import pyarrow.ipc as pi
import pyarrow.parquet as pq
import structlog

ERROR_PRE = "Invalid Arrow stream - "
PARQUET_MARKER = b"PAR1"
ARROW_BEGIN_OF_STREAM_MARKER = b"ARRBOS1"
ARROW_END_OF_STREAM_MARKER = b"ARREOS1"
ARROW_BATCH_MARKER = b"ARRB1"
DEFAULT_MAX_ROWS_PER_BATCH = 1_000_000
APPLICATION_ARROW_STREAM = "application/vnd.apache.arrow.stream"

DEFAULT_MAGIC_BYTES = 2048
SNIFF_SIZE = 1024 * 10


BatchBytesFunction = Callable[[bytes], None]
TableFunction = Callable[[pa.Table], None]

Stream = AsyncGenerator[bytes, None]
FileLike = Union[BinaryIO, IO[Any], io.IOBase]
FileLikes = list[FileLike]
TableSource = Union[str, FileLike]
DataSourceItem = Union[TableSource, pa.Table]
DataSourceItems = Union[DataSourceItem, list[DataSourceItem]]

logger = structlog.get_logger(__name__)


@dataclass
class FileInfo:
    fmt: Optional[str] = None
    error: list[str] = field(default_factory=list)


class StreamState(Enum):
    WAIT_STREAM_BEGIN = "stream-begin"
    AT_STREAM_END = "stream-end"
    WAIT_BATCH = "batch"
    DATA = "data"


def has_method(obj: Any, name: str) -> bool:
    return hasattr(obj, name) and callable(getattr(obj, name))


def is_existing_file_or_dir(item: Any) -> Optional[Path]:
    if isinstance(item, str):
        p = Path(item)

        if p.is_file() or p.is_dir():
            return p

    return None


def is_file_like(obj: Any) -> bool:
    return has_method(obj, "read") and has_method(obj, "write")


def is_text_file(obj: Any) -> bool:
    # Cannot test solely on inheritance because of TemporaryFileWrapper for example
    return is_file_like(obj) and hasattr(obj, "mode") and "b" not in obj.mode


def is_text_buffer(obj: Any) -> bool:
    return isinstance(obj, io.TextIOBase)


def is_text_file_or_buffer(obj: Any) -> bool:
    # Cannot test solely on inheritance because of TemporaryFileWrapper for example
    return is_text_file(obj) or is_text_buffer(obj)


@contextmanager
def save_pos(obj: Any) -> Generator[None, None, None]:
    pos = 0

    if is_file_like(obj):
        pos = obj.tell()

    yield

    if is_file_like(obj):
        obj.seek(pos, os.SEEK_SET)


@contextmanager
def to_start(data: BytesIO, ofs: int = 0) -> Generator[None, None, None]:
    with save_pos(data):
        data.seek(ofs, os.SEEK_SET)
        yield


@contextmanager
def to_end(data: BytesIO) -> Generator[None, None, None]:
    with save_pos(data):
        data.seek(0, os.SEEK_END)
        yield


def available(data: BytesIO) -> int:
    with to_end(data):
        return data.tell()


def try_format(finfo: FileInfo, fmt: str, obj: Any, func: Callable[[Any], Any]) -> None:
    if finfo.fmt:
        return

    with save_pos(obj):
        try:
            func(obj)
            finfo.fmt = fmt
            finfo.error.clear()
        except pa.ArrowInvalid as e:
            finfo.error = [fmt, str(e)]


def guess_file_format(obj: Any) -> FileInfo:
    finfo = FileInfo()

    try_format(finfo, "parquet", obj, lambda obj: pq.ParquetFile(obj))
    try_format(finfo, "csv", obj, lambda obj: pcsv.read_csv(obj))

    if not finfo.fmt:
        finfo.fmt = "unknown"

    return finfo


def flatten(items: Any) -> Iterator[Any]:
    if isinstance(items, list):
        for item in items:
            yield from flatten(item)
    else:
        yield items


class MarkerBase:
    def __init__(self, *, total_size: int) -> None:
        self.total_size = total_size

    def make(self) -> bytes:
        return bytes()

    def can_load(self, data: memoryview) -> bool:
        return len(data) >= self.total_size

    def check_can_load(self, data: memoryview) -> None:
        if not self.can_load(data):
            raise ValueError(
                f"Expected at least {self.total_size} bytes, "
                f"got {len(data)} instead"
            )

    def load(self, data: memoryview) -> memoryview:
        self.check_can_load(data)

        return data

    def try_load(self, data: memoryview) -> bool:
        if self.can_load(data):
            self.load(data)
            return True
        else:
            return False

    def skip_data(self, data: memoryview, size: int) -> memoryview:
        return data[size:]


class Marker(MarkerBase):
    def __init__(self, marker: bytes, *, aux_size: int = 0) -> None:
        self.marker = marker
        self.marker_size = len(self.marker)
        self.aux_size = aux_size
        super().__init__(total_size=self.marker_size + self.aux_size)

    def make(self) -> bytes:
        return super().make() + self.marker

    def identify(self, data: memoryview) -> bool:
        if self.can_load(data):
            return self.check_marker(data, expected=False)

        return False

    def load(self, data: memoryview) -> memoryview:
        # Load "my" parts (marker) and return the rest
        my_data = super().load(data)
        self.check_marker(my_data)

        return self.skip_data(my_data, self.marker_size)

    def check_marker(self, data: memoryview, *, expected: bool = True) -> bool:
        marker = data[0 : self.marker_size]

        if marker != self.marker:
            if expected:
                raise ValueError(
                    f"{ERROR_PRE}"
                    f"Expected '{self.marker!r}', got {bytes(marker)!r} instead"
                )
            else:
                return False

        return True


class MarkerWithSize(Marker):
    size_format = "!I"
    size_packed_size = struct.calcsize(size_format)

    def __init__(self, marker: bytes, size: int = 0) -> None:
        self.size = size
        super().__init__(marker, aux_size=MarkerWithSize.size_packed_size)

    def make(self) -> bytes:
        return super().make() + struct.pack(self.size_format, self.size)

    def load(self, data: memoryview) -> memoryview:
        my_data = super().load(data)
        packed_size = my_data[0 : MarkerWithSize.size_packed_size]

        self.size = struct.unpack(MarkerWithSize.size_format, packed_size)[0]

        return self.skip_data(my_data, MarkerWithSize.size_packed_size)


class ArrowDatasetBuilder:
    def __init__(self) -> None:
        self.inferred_type: Optional[str] = None

    def sniff_csv_data(self, source: TableSource) -> dict[Any, Any]:
        if isinstance(source, (str, Path)):
            with open(source, "rb") as f:
                sample = f.read(SNIFF_SIZE)
        else:
            sample = source.read(SNIFF_SIZE)
            source.seek(0)

        if isinstance(sample, bytes):
            sample = sample.decode(errors="ignore")  # type: ignore[assignment]
        else:
            sample = str(sample)

        try:
            # Python csv module is really struggling on simple cases
            # CleverCSV seems to do a better job
            dialect = clevercsv.Sniffer().sniff(sample)  # type: ignore[arg-type]
        except clevercsv.Error as e:
            raise ValueError(str(e))

        if not dialect:
            raise ValueError("Failed to find CSV format")

        if not dialect.delimiter:
            dialect.delimiter = ","

        return dialect.to_dict()

    def read_table(self, source: TableSource, finfo: FileInfo) -> pa.Table:
        if finfo.fmt == "csv":
            dialect = self.sniff_csv_data(source)
            parse_options = pcsv.ParseOptions(delimiter=dialect["delimiter"])
            return pcsv.read_csv(source, parse_options=parse_options)
        elif finfo.fmt == "parquet":
            return pq.read_table(source)

        if finfo.error:
            detail = finfo.error[1]
        else:
            detail = f"Expected known format, got '{finfo.fmt}' instead"

        raise ValueError(f"{ERROR_PRE}{detail}")

    def to_table(self, source: TableSource) -> pa.Table:
        finfo = guess_file_format(source)
        self.inferred_type = finfo.fmt

        logger.info(f"dataset item guessed format is {finfo.fmt}")

        return self.read_table(source, finfo)

    def to_source(self, item: Any) -> Union[str, pa.Table]:
        if p := is_existing_file_or_dir(item):
            if p.is_file():
                return self.to_table(item)
            else:
                return item
        elif is_text_file(item):
            # Let pyarrow open the file itself (in binary mode)
            return self.to_table(item.name)
        elif isinstance(item, io.IOBase):
            if is_text_file_or_buffer(item):
                item = io.BytesIO(item.read().encode())

            return self.to_table(item)
        elif isinstance(item, pd.DataFrame):
            return pa.Table.from_pandas(item)
        else:
            raise TypeError(f"Unsupported dataset source {type(item)}")

    def to_dataset(self, items: DataSourceItems) -> tuple[ds.Dataset, Optional[str]]:
        self.inferred_type = None
        sources = [self.to_source(s) for s in flatten(items)]

        return ds.dataset(sources), self.inferred_type


class ArrowStreamReader:
    def __init__(
        self,
        *,
        batch_bytes_func: Optional[BatchBytesFunction] = None,
        table_func: Optional[TableFunction] = None,
    ) -> None:
        self.batch_bytes_func = batch_bytes_func
        self.table_func = table_func
        self.func_stack: list[Any] = []
        self.batch = bytes()
        self.data = bytearray()
        self.stream_header = MarkerWithSize(ARROW_BEGIN_OF_STREAM_MARKER)
        self.stream_footer = MarkerWithSize(ARROW_END_OF_STREAM_MARKER)
        self.batch_header = MarkerWithSize(ARROW_BATCH_MARKER)
        self.state = StreamState.WAIT_STREAM_BEGIN
        self.nb_batches = 0
        self.total_rows = None

    def __enter__(self) -> Any:
        return self

    def __exit__(self, *args: Any) -> Any:
        pass

    def write_parquet(
        self,
        where: Union[str, IO[bytes]],
        iter_bytes: Iterator[bytes],
    ) -> None:
        writer: Optional[pq.ParquetWriter] = None

        logger.info(f"WP: file is {where}")

        def write_table(table: pa.Table) -> None:
            # In case you wonder: writer is created here on the first run because ParquetWriter
            # needs a table schema, hence a table, which we only have on the first call when
            # the first table is produced (while streaming from server)
            nonlocal writer

            if not writer:
                writer = pq.ParquetWriter(where, table.schema)

            logger.info(f"WP: {where=} {table.num_rows=}")
            writer.write_table(table)

        self.process_bytes(iter_bytes, table_func=write_table)

        if writer:
            writer.close()

        logger.info(f"WP: {where=} closing")

    async def process_stream(
        self,
        stream: Stream,
        *,
        batch_bytes_func: Optional[BatchBytesFunction] = None,
        table_func: Optional[TableFunction] = None,
    ) -> None:
        self.process_begin(batch_bytes_func=batch_bytes_func, table_func=table_func)

        async for chunk in stream:
            self.process_chunk(chunk)

        self.process_end()

    def process_bytes(
        self,
        iter_bytes: Iterator[bytes],
        *,
        batch_bytes_func: Optional[BatchBytesFunction] = None,
        table_func: Optional[TableFunction] = None,
    ) -> None:
        self.process_begin(batch_bytes_func=batch_bytes_func, table_func=table_func)

        for chunk in iter_bytes:
            self.process_chunk(chunk)

        self.process_end()

    def process_begin(
        self,
        *,
        batch_bytes_func: Optional[BatchBytesFunction] = None,
        table_func: Optional[TableFunction] = None,
    ) -> None:
        self.push_funcs()
        self.batch_bytes_func = batch_bytes_func
        self.table_func = table_func

    def process_end(self) -> None:
        self.update_state()
        self.check_stream_started()
        self.flush()
        self.check_at_stream_end()
        self.check_no_remaining_data()
        self.check_nb_batches()
        self.pop_funcs()

    def process_chunk(self, chunk: bytes) -> None:
        if len(chunk) == 0:
            return

        if self.state == StreamState.AT_STREAM_END:
            raise ValueError("Receiving data after stream end")

        self.data.extend(chunk)
        self.update_state()

    def update_state(self) -> None:
        while True:
            prev_state = self.state

            if self.state == StreamState.WAIT_STREAM_BEGIN:
                self.try_load_header(self.stream_header, StreamState.WAIT_BATCH)
                self.total_rows = self.stream_header.size  # type: ignore[assignment]
            elif self.state == StreamState.WAIT_BATCH:
                self.try_load_header(self.batch_header, StreamState.DATA)
            elif self.state == StreamState.DATA:
                self.flush()

            if self.state == prev_state:
                # No progress in this loop, need more data
                break

    def try_load_header(self, header: Marker, new_state: StreamState) -> None:
        data = memoryview(self.data)

        if self.stream_footer.identify(data):
            self.stream_footer.load(data)
            # End of stream
            self.skip_marker_and_set_state(
                self.stream_footer, StreamState.AT_STREAM_END
            )
        elif header.try_load(data):
            self.skip_marker_and_set_state(header, new_state)

    def skip_marker_and_set_state(
        self, marker: MarkerBase, new_state: StreamState
    ) -> None:
        self.skip_data(marker.total_size)
        self.state = new_state

    def flush(self) -> None:
        if not self.is_batch_complete() or self.no_data():
            return

        # Read batch bytes
        self.batch = self.extract_data(self.batch_header.size)
        self.nb_batches += 1

        # Keep remaining data from next batch
        self.skip_data(self.batch_header.size)

        self.call_funcs()

        self.set_wait_batch()

    def is_batch_complete(self) -> Any:
        return len(self.data) >= self.batch_header.size

    def set_wait_batch(self) -> None:
        self.state = StreamState.WAIT_BATCH
        self.batch_header.size = 0

    def push_funcs(self) -> None:
        self.func_stack.append([self.batch_bytes_func, self.table_func])

    def pop_funcs(self) -> None:
        self.batch_bytes_func, self.table_func = self.func_stack.pop()

    def call_funcs(self) -> None:
        if self.batch_bytes_func:
            self.batch_bytes_func(self.batch)

        if self.table_func:
            reader = pi.RecordBatchStreamReader(self.batch)
            nb_args = len(inspect.signature(self.table_func).parameters)
            if nb_args == 2:
                self.table_func(  # type: ignore[call-arg]
                    reader.read_all(),
                    self.total_rows,
                )
            elif nb_args == 1:
                self.table_func(reader.read_all())
            else:
                raise ValueError(
                    f"Expected 1 or 2 arguments for table_func, got {nb_args}"
                )

    def skip_data(self, size: int) -> None:
        self.data = self.data[size:]

    def extract_data(self, size: int) -> bytes:
        return bytes(self.data[0:size])

    def no_data(self) -> bool:
        return len(self.data) == 0

    def check_no_remaining_data(self) -> None:
        remaining = len(self.data)

        if remaining:
            raise ValueError(f"Expected no remaining data, got {remaining} bytes")

    def check_nb_batches(self) -> None:
        if self.nb_batches != self.stream_footer.size:
            raise ValueError(
                f"Expected {self.stream_footer.size} batches, "
                f"got {self.nb_batches} instead"
            )

    def check_stream_started(self) -> None:
        if self.state == StreamState.WAIT_STREAM_BEGIN:
            raise ValueError(f"Expected a stream, got {bytes(self.data)!r} instead")

    def check_at_stream_end(self) -> None:
        if self.state != StreamState.AT_STREAM_END:
            raise ValueError(
                "Expected to be at stream end, " f"got {self.state.value} instead"
            )


class ArrowStreamWriter:
    def __init__(
        self, ds: ds.Dataset, nb_rows_per_batch: int = DEFAULT_MAX_ROWS_PER_BATCH
    ) -> None:
        self.ds = ds
        self.total_rows = ds.count_rows()
        self.nb_rows_per_batch = nb_rows_per_batch
        self.batch_header = MarkerWithSize(ARROW_BATCH_MARKER)
        self.batch: Optional[bytes] = None
        self.nb_batches = 0
        self.reset_writer()

    def __enter__(self) -> Any:
        return self

    def __exit__(self, *args: Any) -> Any:
        pass

    def iter_core(self) -> Iterator[bytes]:
        yield MarkerWithSize(ARROW_BEGIN_OF_STREAM_MARKER, size=self.total_rows).make()

        for batch in self.ds.to_batches(batch_size=self.nb_rows_per_batch):
            yield from self.process_batch(batch)

        yield from self.flush()

        yield MarkerWithSize(ARROW_END_OF_STREAM_MARKER, self.nb_batches).make()

    async def stream(self) -> Stream:
        for v in self.iter_core():
            yield v

    def __iter__(self) -> Iterator[bytes]:
        yield from self.iter_core()

    def process_batch(self, batch: pa.RecordBatch) -> Iterator[bytes]:
        if self.should_flush(batch):
            yield from self.flush()

        self.writer.write_batch(batch)
        self.row_count += batch.num_rows

    def should_flush(self, batch: pa.RecordBatch) -> Any:
        return self.row_count + batch.num_rows > self.nb_rows_per_batch

    def flush(self) -> Iterator[bytes]:
        self.close_writer()
        self.reset_writer()
        self.nb_batches += 1

        yield from self.get_buffers()

    def get_buffers(self) -> Iterator[bytes]:
        if self.batch:
            self.batch_header.size = len(self.batch)

            yield self.batch_header.make()
            yield self.batch

            self.batch = None

    def close_writer(self) -> None:
        self.writer.close()
        buffer = self.buf.getvalue()
        self.batch = buffer if len(buffer) else None

    def reset_writer(self) -> None:
        self.buf = io.BytesIO()
        self.writer = pi.RecordBatchStreamWriter(self.buf, self.ds.schema)
        self.row_count = 0



from typing import List

import numpy as np
import pandas as pd

from avatars.lib.saferound import saferound


class ProportionProcessor:
    """Processor to express numeric variables as a proportion of another variable.

    By this transformation, we keep the addition and subtraction relations
    such as variable_1 = variable_2 + variable_3.

    Arguments
    ---------
        variable_names:
            variables to transform
        reference:
            the variable of reference

    Keyword Arguments
    -----------------
        sum_to_one:
            set to True to ensure the sum of the variables sum to 1 once transformed.
            default: True
        decimal_count:
            the number of decimals postprocessed variables should have

    Examples
    --------
    >>> df =  pd.DataFrame(
    ...        {
    ...            "variable_1": [100, 10],
    ...            "variable_2": [10, 10],
    ...            "variable_3": [90, 30],
    ...        }
    ...    )
    >>> processor = ProportionProcessor(
    ...    variable_names=["variable_2", "variable_3"],
    ...    reference="variable_1",
    ... )
    >>> processor.preprocess(df=df)
       variable_1  variable_2  variable_3
    0         100        0.10        0.90
    1          10        0.25        0.75

    This processor allows you to transform some variable as a proportion of another variable.
    By default, the processor enforces the proportion of `variable_names` to be equal to 1.

    >>> avatar = pd.DataFrame(
    ...        {
    ...            "variable_1": [60, 15],
    ...            "variable_2": [0.15, 0.88],
    ...            "variable_3": [0.18, 0.77],
    ...        }
    ...    )
    >>> avatar
       variable_1  variable_2  variable_3
    0          60        0.15        0.18
    1          15        0.88        0.77

    Then the postprocess allows you to get the original variable unit.

    >>> processor.postprocess(df, avatar)
       variable_1  variable_2  variable_3
    0          60        27.3        32.7
    1          15         8.0         7.0

    By this, we keep the mathematical relation variable_1 = variable_2 + variable_3

    with `sum_to_one=False`

    >>> processor = ProportionProcessor(
    ...    variable_names=["variable_2", "variable_3"],
    ...    reference="variable_1",
    ...    sum_to_one=False,
    ... )
    >>> processor.preprocess(df=df)
       variable_1  variable_2  variable_3
    0         100         0.1         0.9
    1          10         1.0         3.0
    >>> avatar = pd.DataFrame(
    ...        {
    ...            "variable_1": [60, 15],
    ...            "variable_2": [0.15, 0.88],
    ...            "variable_3": [1.5, 2.8],
    ...        }
    ...    )
    >>> avatar
       variable_1  variable_2  variable_3
    0          60        0.15         1.5
    1          15        0.88         2.8
    >>> processor.postprocess(df, avatar)
       variable_1  variable_2  variable_3
    0          60         9.0        90.0
    1          15        13.2        42.0
    """

    def __init__(
        self,
        variable_names: List[str],
        reference: str,
        *,
        sum_to_one: bool = True,
        decimal_count: int = 1,
    ):
        self.variable_names = variable_names
        self.reference = reference
        self.sum_to_one = sum_to_one
        self.decimal_count = decimal_count

    def preprocess(self, df: pd.DataFrame) -> pd.DataFrame:
        """Transform numeric variables into proportion of another variable.

        If some values for the variables to transform are set to nan, they will be transformed
        into nan and will be considered as a 0% proportion of the reference when transforming
        values of other variables.

        Arguments
        ---------
            df: dataframe to transform

        Returns
        -------
            DataFrame: a dataframe with the transformed version of wanted columns
        """
        col_order = df.columns
        df = df.copy()
        if self.reference not in df.columns.values:
            raise ValueError(
                "variable_name",
                f"variable {self.reference} cannot be found in the dataframe variables",
            )
        if self.sum_to_one:
            sub_df = df[self.variable_names].div(
                df[self.variable_names].sum(axis=1), axis=0
            )
        else:
            sub_df = df[self.variable_names].div(df[self.reference], axis=0)

        # Ensure that target variables that were set to zero remain at zero
        for variable in self.variable_names:
            zero_indices = df[df[variable] == 0].index
            if len(zero_indices) > 0:
                sub_df.loc[zero_indices, variable] = 0

        df = df.drop(columns=self.variable_names)
        df = pd.concat([df, sub_df], axis=1)[col_order]
        return df

    def postprocess(self, source: pd.DataFrame, dest: pd.DataFrame) -> pd.DataFrame:
        """Transform proportion of another variable into an absolute numeric value.

        Arguments
        ---------
            source: not used
            dest: dataframe to transform

        Returns
        -------
            DataFrame: a dataframe with the transformed version of wanted columns
        """
        col_order = dest.columns
        dest = dest.copy()
        if self.reference not in dest.columns.values:
            raise ValueError(
                "variable_name",
                f"variable {self.reference} cannot be found in the dataframe variables",
            )
        if self.sum_to_one:
            sub_df = (
                dest[self.variable_names]
                .mul(dest[self.reference], axis=0)
                .div(dest[self.variable_names].sum(axis=1), axis=0)
            )
        else:
            sub_df = dest[self.variable_names].mul(dest[self.reference], axis=0)
        dest = dest.drop(columns=self.variable_names)
        dest = pd.concat([dest, sub_df], axis=1)[col_order]

        # Perform rounding of the postprocess variable to the expected number of decimals.
        # We use saferounding here to force the sum of rounded variables to remain unchanged.
        if not self.sum_to_one:
            return dest

        for i, row in enumerate(dest[self.variable_names].values):
            if not np.any(np.isnan(row)):
                dest.loc[i, self.variable_names] = saferound(
                    row.tolist(), self.decimal_count
                )

        return dest




from typing import Dict, List, Tuple

import numpy as np
import pandas as pd


class GeolocationNormalizationProcessor:
    """Processor to normalize longitude values for different latitude.

    Use of this processor is recommended on data with latitude and longitude variables where the
    overall area covered has regions with no points (e.g. lake, forests etc ...)
    or non-squared bounds (e.g. country borders) and it is required to retain this lack of points
    in the anonymized data.

    Keyword Arguments
    -----------------
        latitude_variable:
            latitude variable name
        longitude_variable:
            longitude variable name
        n_reference_lat:
            number of discretized reference latitudes. A high number of references will yield a
            higher fidelity in the latitude dimension
        n_bins:
            number of discretized longitude bins at each reference latitude.  A high number of
            bins will yield a higher fidelity in the longitude dimension.

    Examples
    --------
    >>> import numpy as np
    >>> df = pd.DataFrame(
    ...    {
    ...        'lat': [49.1, 49.2, 49.3, 49.3],
    ...        'lon': [3.21, 3.19, 3.11, 3.18]
    ...    }
    ... )
    >>> df
        lat   lon
    0  49.1  3.21
    1  49.2  3.19
    2  49.3  3.11
    3  49.3  3.18
    >>> processor = GeolocationNormalizationProcessor(
    ...    latitude_variable='lat',
    ...    longitude_variable='lon',
    ...    n_reference_lat=3,
    ...    n_bins=5
    ... )
    >>> processed = processor.preprocess(df)
    >>> processed
        lat  lon
    0  49.1  0.5
    1  49.2  0.5
    2  49.3  0.0
    3  49.3  1.0

    The pre process expresses one of the coordinate dimension (i.e. lon) between 0 and 1.
    The range of values is different for different lat.

    >>> processor.postprocess(source=df, dest=processed)
        lat   lon
    0  49.1  3.21
    1  49.2  3.19
    2  49.3  3.11
    3  49.3  3.18

    The post process re-express the coordinates in the same original ranges.

    """

    def __init__(
        self,
        *,
        latitude_variable: str,
        longitude_variable: str,
        n_reference_lat: int,
        n_bins: int
    ):
        self.latitude_variable = latitude_variable
        self.longitude_variable = longitude_variable
        self.n_reference_lat = n_reference_lat
        self.n_bins = n_bins
        self.model: Dict[float, Dict[str, List[Tuple[float, float]]]] = {}


    def _compute_references(self, df: pd.DataFrame) -> pd.DataFrame:
        """Define reference latitudes and associate original points to each."""

        # Pick n lats along the lat range
        reference_lats = np.linspace(min(df['lat']), max(df['lat']), self.n_reference_lat)

        # align points to reference lats
        delta_reference = (max(df['lat'])-min(df['lat'])) / (self.n_reference_lat-1)
        # points_to_reference_tolerance: defines which points are mapped to each lat reference.
        # points_to_reference_tolerance = 0 -> only points with the exact lat are mapped.
        # points_to_reference_tolerance = 0.2 -> points that are within 20% of the distance
        #    between two reference lats are mapped to each reference
        # Using points_to_reference_tolerance >= 0.5 ensures all points are mapped to a
        #    reference lat. This ensure transformed values are within [0, 1] bounds.
        points_to_reference_tolerance = 0.5
        tol = points_to_reference_tolerance * delta_reference

        ref_dfs = []  # copy of original df where lat is changed to the associated ref lat
        for ref_lat in reference_lats:
            # get points matching the ref with the given tolerance tol
            tmp_df  = df[(df['lat'] >= (ref_lat - tol)) & (df['lat'] <= (ref_lat + tol))][['lat', 'lon']].copy()
            tmp_df['lat'] = ref_lat # set lat as ref_lat
            ref_dfs.append(tmp_df)
        ref_df = pd.concat(ref_dfs).reset_index(drop=True)

        return ref_df

    def _compute_ranges(self, ref_df: pd.DataFrame, nbins: int) -> None:
        """Calculate ranges of lon where there are points for each reference lat"""

        # "Cut" values of each reference lat in `nbins` bins. A bin represent a portion of the lon
        # covered at that lat
        for ref_lat in ref_df['lat'].unique():

            range_min = min(ref_df[ref_df['lat'] == ref_lat]['lon'])
            range_max = max(ref_df[ref_df['lat'] == ref_lat]['lon'])

            # count number of points in each of the bin (i.e. each of the bin)
            pmf1, bin_edges = np.histogram(
                    ref_df[ref_df['lat'] == ref_lat]['lon'],
                    bins=nbins,
                    range=(range_min, range_max),
                    density=True,
                )

            # only keep ranges where there are points
            non_zero_bin_starts = np.where(np.array(pmf1) > 0)[0]
            # and store the start and end lon of each range in a list of tuples
            ranges = [(bin_edges[bin_start], bin_edges[bin_start+1]) for bin_start in non_zero_bin_starts]
            n_ranges = len(ranges)
            # also store the proportion of total lon covered by each range. Range proportions
            # should sum to 1 for each reference lat.
            ranges_proportions = [(i/n_ranges, i/n_ranges + 1/n_ranges) for i in range(n_ranges)]

            self.model[ref_lat] = {'ranges': ranges, 'ranges_proportions': ranges_proportions}

    def _transform_one_point(self, row: pd.Series) -> pd.Series:
        # get closest reference lat for point to transform
        refs = list(self.model.keys())
        closest_ref = min(refs, key=lambda x:abs(x-row['lat']))

        # compute proportion value based on closest reference lat using the lon of
        # the point to. The proportion value is based on the range in which the original lon falls
        # and is a linear estimation calculated from start and end of the range
        selected_i = -1
        for i, r in enumerate(self.model[closest_ref]['ranges']):
            if r[0] <= row['lon'] < r[1]:
                selected_i = i
                break

        min_val = self.model[closest_ref]['ranges'][selected_i][0]
        max_val = self.model[closest_ref]['ranges'][selected_i][1]
        proportion_of_val = (row['lon'] - min_val)/(max_val - min_val)

        min_val = self.model[closest_ref]['ranges_proportions'][selected_i][0]
        max_val = self.model[closest_ref]['ranges_proportions'][selected_i][1]
        transformed_val = min_val + proportion_of_val * (max_val - min_val)

        x_transformed = row.copy()
        x_transformed['lon'] = transformed_val

        return x_transformed

    def _inv_transform_one_point(self, row: pd.Series) -> pd.Series:
        # inverse transform follows the same logic as the transform (see _transform_one_point)
        refs = list(self.model.keys())
        closest_ref = min(refs, key=lambda x:abs(x-row['lat']))

        selected_i = -1
        for i, r in enumerate(self.model[closest_ref]['ranges_proportions']):
            if r[0] <= row['lon'] < r[1]:
                selected_i = i
                break

        min_val = self.model[closest_ref]['ranges_proportions'][selected_i][0]
        max_val = self.model[closest_ref]['ranges_proportions'][selected_i][1]
        proportion_of_val = (row['lon'] - min_val)/(max_val - min_val)

        min_val = self.model[closest_ref]['ranges'][selected_i][0]
        max_val = self.model[closest_ref]['ranges'][selected_i][1]
        transformed_val = min_val + proportion_of_val * (max_val - min_val)

        x_inv_transformed = row.copy()
        x_inv_transformed['lon'] = transformed_val

        return x_inv_transformed


    def _fit(self, df: pd.DataFrame) -> None:
        self.ref_df = self._compute_references(df)
        self._compute_ranges(self.ref_df, self.n_bins)


    def preprocess(self, df: pd.DataFrame) -> pd.DataFrame:
        self._fit(df)

        transformed_tmp = []
        for i in range(len(df)):
            x = df.iloc[i]
            x_transformed = self._transform_one_point(x)
            transformed_tmp.append(x_transformed)
        df_transformed = pd.concat(transformed_tmp, axis=1).T

        return df_transformed


    def postprocess(self, source: pd.DataFrame, dest: pd.DataFrame) -> pd.DataFrame:
        inv_transformed_tmp = []
        for i in range(len(dest)):
            x = dest.iloc[i]
            x_transformed = self._inv_transform_one_point(x)
            inv_transformed_tmp.append(x_transformed)
        df_transformed = pd.concat(inv_transformed_tmp, axis=1).T
        return df_transformed



import math
from typing import Dict, List, Optional, Tuple

import numpy as np
import pandas as pd


def get_split_for_batch(
    df: pd.DataFrame,
    row_limit: int,
    seed: Optional[int] = None,
) -> Tuple[pd.DataFrame, List[pd.DataFrame]]:
    """Create batches from a dataframe with the format (training, splits).


    Arguments
    ---------
        df:
            dataframe to split into batch
        row_limit:
            maximum number of rows per batch, this parameter determines the
            number of batches that will be created

    Keyword Arguments
    -----------------
        seed:
            seed of the random split for the batch

    Returns
    -------
        training:
            the training batch that will be used to fit the anonymization
        splits:
            all other batches

    Examples
    --------
    >>> df = pd.DataFrame(
    ... data={
    ...     "a": [1, 3, 5, 1, 3, 5, 1, 3, 5, 1, 3, 5],
    ...     "b": ["a", "b", "a", "a", "b", "a", "a", "b", "a", "a", "b", "a"],
    ...     }
    ... )
    >>> training, splits = get_split_for_batch(df, row_limit=6, seed=42)
    >>> training
        a  b
    0   1  a
    1   3  b
    10  3  b
    3   1  a
    7   3  b
    2   5  a
    >>> splits
    [    a  b
    9   1  a
    4   3  b
    11  5  a
    6   1  a
    5   5  a
    8   5  a]
    """
    number_of_split = math.ceil(len(df) / row_limit)
    categorical_columns = df.select_dtypes(include=["object", "category"]).columns

    index = get_index_without_duplicated_modalities(df[categorical_columns])
    first_occurrence = df.loc[index, :]
    if len(first_occurrence) > row_limit * 0.5:
        raise ValueError(
            "You have too much variability in your categorical data. "
            "You can increase your `row_limit`, reduce your categorical variability, or build "
            "your own batch pipeline."
        )

    df_without_first = df.drop(index=first_occurrence.index)

    shuffled = df_without_first.sample(frac=1, random_state=seed)

    all_df = pd.concat([first_occurrence, shuffled])
    splits = np.array_split(all_df, number_of_split)
    splits = [pd.DataFrame(split) for split in splits]
    training = splits.pop(0)

    return training, splits


def get_index_without_duplicated_modalities(df: pd.DataFrame) -> List[int]:
    """Get a small number of indexes to subset a dataframe by selecting all modalities."""
    already_seen: Dict[str, Dict[str, bool]] = {col: {} for col in df.columns}
    df_array = df.to_numpy()
    index_to_keep = []
    # We travel all the rows and the columns
    # for each case, we check if we have already have seen this value
    # if the value was never seen, we save it into a dictionary
    # and save it's index
    for i in range(df_array.shape[0]):
        should_keep_row = False
        for j in range(df_array.shape[1]):
            column = df.columns[j]
            try:
                already_seen[column][df_array[i, j]]
            except KeyError:
                already_seen[column][df_array[i, j]] = True
                should_keep_row = True
        if should_keep_row:
            index_to_keep.append(i)
    return index_to_keep



from typing import List, Optional

import pandas as pd
import pandas.testing as pd_testing
import pytest

from avatars.processors import ExpectedMeanProcessor


@pytest.fixture
def original_df() -> pd.DataFrame:
    return pd.DataFrame(
        {
            "variable_1": ["a", "a", "b", "b", "b"],
            "variable_2": [1, 3, 3, 4, 6],
            "variable_3": [3, 3, 1, 2, 3],
            "variable_4": [10, 9, 30, 40, 50],
        }
    )


@pytest.fixture
def raw_avatars_df() -> pd.DataFrame:
    return pd.DataFrame(
        {
            "variable_1": ["a", "a", "b", "b", "b"],
            "variable_2": [1, 3, 3, 4, 6],
            "variable_3": [5, 3, 2, 2, 6],
            "variable_4": [8, 11, 20, 42, 55],
        }
    )


def test_preprocess_with_groupby(original_df: pd.DataFrame) -> None:
    # Verify that the calculated properties of the target variables are correct.
    # NB: pandas uses Bessel's correction to calculate std, while numpy does not. This brings
    # different std results on very small datasets. The expected values in the test is based on
    # the use of Bessel's correction.
    processor = ExpectedMeanProcessor(
        target_variables=["variable_3", "variable_4"],
        groupby_variables=["variable_1"],
        same_std=False,
    )
    processor.preprocess(df=original_df)
    expected = pd.DataFrame(
        {
            "variable_1": ["a", "b"],
            "variable_3mean": [3.0, 2.0],
            "variable_3std": [0.0, 1.0],
            "variable_4mean": [9.5, 40.0],
            "variable_4std": [0.707107, 10.0],
        }
    )

    pd_testing.assert_frame_equal(processor.properties_df, expected)


def test_postrocess_with_groupby(
    original_df: pd.DataFrame, raw_avatars_df: pd.DataFrame
) -> None:
    # Verify that the postprocessed dataframe is correct when using only
    # the mean to post-process the data.
    processor = ExpectedMeanProcessor(
        target_variables=["variable_3", "variable_4"],
        groupby_variables=["variable_1"],
        same_std=False,
    )
    processor.preprocess(df=original_df)
    fixed_avatars_df = processor.postprocess(source=original_df, dest=raw_avatars_df)
    expected = pd.DataFrame(
        {
            "variable_1": ["a", "a", "b", "b", "b"],
            "variable_2": [1, 3, 3, 4, 6],
            "variable_3": [4, 2, 0.66667, 0.66667, 4.66667],
            "variable_4": [8, 11, 21, 43, 56],
        }
    )

    pd_testing.assert_frame_equal(fixed_avatars_df, expected, check_dtype=False)


@pytest.mark.parametrize(
    "groupby_variables, expected",
    [
        (
            ["variable_1"],
            pd.DataFrame(
                {
                    "variable_1": ["a", "a", "b", "b", "b"],
                    "variable_2": [1, 3, 3, 4, 6],
                    "variable_3": [3, 3, 1.42265, 1.42265, 3.154701],
                    "variable_4": [9, 10, 29.260565, 41.6957, 49.043735],
                }
            ),
        ),
        (
            None,
            pd.DataFrame(
                {
                    "variable_1": ["a", "a", "b", "b", "b"],
                    "variable_2": [1, 3, 3, 4, 6],
                    "variable_3": [3.09, 2.10, 1.61, 1.61, 3.58],
                    "variable_4": [10.78, 13.44, 21.42, 40.92, 52.45],
                }
            ),
        ),
    ],
)
def test_postprocess_same_std(
    original_df: pd.DataFrame,
    raw_avatars_df: pd.DataFrame,
    groupby_variables: Optional[List[str]],
    expected: pd.DataFrame,
) -> None:
    # Verify that the postprocessed dataframe is correct when none of the
    # variables to transform has a std equal to zero.
    processor = ExpectedMeanProcessor(
        target_variables=["variable_3", "variable_4"],
        groupby_variables=groupby_variables,
        same_std=True,
    )
    processor.preprocess(df=original_df)
    fixed_avatars_df = processor.postprocess(source=original_df, dest=raw_avatars_df)

    pd_testing.assert_frame_equal(fixed_avatars_df, expected, atol=0.01)


@pytest.mark.filterwarnings("ignore:.*std of 0*")
def test_postrocess_same_std_with_zero_std() -> None:
    # Verify that the postprocessed dataframe is correct when one of the
    # variables to transform has a std equal to zero.
    original_df = pd.DataFrame(
        {
            "variable_1": ["a", "a", "b", "b", "b"],
            "variable_2": [1, 3, 3, 4, 6],
            "variable_3": [3, 4, 1, 2, 3],
            "variable_4": [10, 9, 30, 40, 50],
        }
    )
    processor = ExpectedMeanProcessor(
        target_variables=["variable_3", "variable_4"],
        groupby_variables=["variable_1"],
        same_std=True,
    )
    processor.preprocess(df=original_df)
    raw_avatars_df = pd.DataFrame(
        {
            "variable_1": ["a", "a", "b", "b", "b"],
            "variable_2": [1, 3, 3, 4, 6],
            "variable_3": [5, 3, 2, 2, 6],
            "variable_4": [8, 8, 20, 42, 55],
        }
    )
    fixed_avatars_df = processor.postprocess(source=original_df, dest=raw_avatars_df)
    expected = pd.DataFrame(
        {
            "variable_1": ["a", "a", "b", "b", "b"],
            "variable_2": [1, 3, 3, 4, 6],
            "variable_3": [4, 3, 1.42265, 1.42265, 3.154701],
            "variable_4": [9.5, 9.5, 21, 43, 56],
        }
    )

    pd_testing.assert_frame_equal(fixed_avatars_df, expected)


def test_preprocess_no_groupby(original_df: pd.DataFrame) -> None:
    # Verify that the calculated properties of the target variables are correct when using no
    # groupby.
    # NB: pandas uses Bessel's correction to calculate std, while numpy does not. This brings
    # different std results on very small datasets. The expected values in the test is based on
    # the use of Bessel's correction.
    processor = ExpectedMeanProcessor(
        target_variables=["variable_3", "variable_4"],
        groupby_variables=None,
        same_std=True,
    )
    processor.preprocess(df=original_df)
    expected = pd.DataFrame(
        {
            processor.nogroup_name: [processor.nogroup_value],
            "variable_3mean": [2.4],
            "variable_3std": [0.89],
            "variable_4mean": [27.8],
            "variable_4std": [18.14],
        }
    )

    pd_testing.assert_frame_equal(processor.properties_df, expected, atol=0.01)



import numpy as np
import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from avatars.processors import InterRecordCumulatedDifferenceProcessor


@pytest.fixture
def df_with_cumulated() -> pd.DataFrame:
    # Values in the test data is purposely not ordered by id nor value
    df = pd.DataFrame(
        {
            "id": [1, 2, 1, 1, 2, 2],
            "value": [1025, 20042, 1000, 1130, 20000, 20040],
        }
    )
    return df


@pytest.fixture
def preprocessed_df_with_cumulated() -> pd.DataFrame:
    df = pd.DataFrame(
        {
            "id": [1, 2, 1, 1, 2, 2],
            "first_value": [1000, 20000, 1000, 1000, 20000, 20000],
            "value_difference": [25, 2, 0, 105, 0, 40],
        }
    )
    return df


def test_preprocess(
    df_with_cumulated: pd.DataFrame, preprocessed_df_with_cumulated: pd.DataFrame
) -> None:
    "Verify that the preprocess is correct."
    processor = InterRecordCumulatedDifferenceProcessor(
        id_variable="id",
        target_variable="value",
        new_first_variable_name="first_value",
        new_difference_variable_name="value_difference",
    )

    processed_df = processor.preprocess(df_with_cumulated)

    assert_frame_equal(processed_df, preprocessed_df_with_cumulated, check_dtype=False)


def test_postprocess_with_keep_order(
    df_with_cumulated: pd.DataFrame, preprocessed_df_with_cumulated: pd.DataFrame
) -> None:
    "Verify that the postprocess is correct."
    processor = InterRecordCumulatedDifferenceProcessor(
        id_variable="id",
        target_variable="value",
        new_first_variable_name="first_value",
        new_difference_variable_name="value_difference",
        keep_record_order=True,
    )

    postprocessed_df = processor.postprocess(
        df_with_cumulated, preprocessed_df_with_cumulated
    )
    assert_frame_equal(postprocessed_df, df_with_cumulated, check_dtype=False)


def test_postprocess_without_keep_order(
    df_with_cumulated: pd.DataFrame, preprocessed_df_with_cumulated: pd.DataFrame
) -> None:
    "Verify that the postprocess is correct."
    processor = InterRecordCumulatedDifferenceProcessor(
        id_variable="id",
        target_variable="value",
        new_first_variable_name="first_value",
        new_difference_variable_name="value_difference",
        keep_record_order=False,
    )

    # make indices of processed_df different than those of df_with_cumulated
    preprocessed_df_with_cumulated.index = range(
        10, 10 + len(preprocessed_df_with_cumulated), 1
    )

    # records should be decoded following row order
    expected = pd.DataFrame(
        {
            "id": [1, 2, 1, 1, 2, 2],
            "value": [1025, 20002, 1025, 1130, 20002, 20042],
        }
    )
    postprocessed_df = processor.postprocess(
        df_with_cumulated, preprocessed_df_with_cumulated
    ).reset_index(drop=True)

    assert_frame_equal(postprocessed_df, expected, check_dtype=False)


def test_preprocess_raises_error_when_missing_ids(
    df_with_cumulated: pd.DataFrame,
) -> None:
    df_with_cumulated.loc[0, "id"] = np.nan
    processor = InterRecordCumulatedDifferenceProcessor(
        id_variable="id",
        target_variable="value",
        new_first_variable_name="first_value",
        new_difference_variable_name="value_difference",
    )

    with pytest.raises(ValueError, match="Expected no missing values for id variable"):
        processor.preprocess(df_with_cumulated)


def test_preprocess_raises_error_when_missing_target(
    df_with_cumulated: pd.DataFrame,
) -> None:
    df_with_cumulated.loc[2, "value"] = np.nan
    processor = InterRecordCumulatedDifferenceProcessor(
        id_variable="id",
        target_variable="value",
        new_first_variable_name="first_value",
        new_difference_variable_name="value_difference",
    )

    with pytest.raises(
        ValueError, match="Expected no missing values for target variable"
    ):
        processor.preprocess(df_with_cumulated)


def test_postprocess_raises_error_when_missing_source_ids(
    df_with_cumulated: pd.DataFrame, preprocessed_df_with_cumulated: pd.DataFrame
) -> None:
    df_with_cumulated.loc[0, "id"] = np.nan

    processor = InterRecordCumulatedDifferenceProcessor(
        id_variable="id",
        target_variable="value",
        new_first_variable_name="first_value",
        new_difference_variable_name="value_difference",
    )

    with pytest.raises(
        ValueError, match="Expected no missing values for id variable in source"
    ):
        processor.postprocess(df_with_cumulated, preprocessed_df_with_cumulated)


def test_preprocess_raises_error_when_missing_source_target(
    df_with_cumulated: pd.DataFrame, preprocessed_df_with_cumulated: pd.DataFrame
) -> None:
    df_with_cumulated.loc[2, "value"] = np.nan

    processor = InterRecordCumulatedDifferenceProcessor(
        id_variable="id",
        target_variable="value",
        new_first_variable_name="first_value",
        new_difference_variable_name="value_difference",
    )

    with pytest.raises(
        ValueError, match="Expected no missing values for target variable in source"
    ):
        processor.postprocess(df_with_cumulated, preprocessed_df_with_cumulated)


def test_postprocess_raises_error_when_missing_dest_first_var(
    df_with_cumulated: pd.DataFrame, preprocessed_df_with_cumulated: pd.DataFrame
) -> None:
    preprocessed_df_with_cumulated.loc[4, "first_value"] = np.nan

    processor = InterRecordCumulatedDifferenceProcessor(
        id_variable="id",
        target_variable="value",
        new_first_variable_name="first_value",
        new_difference_variable_name="value_difference",
    )

    with pytest.raises(
        ValueError, match="Expected no missing values for `new_first_variable_name`"
    ):
        processor.postprocess(df_with_cumulated, preprocessed_df_with_cumulated)


def test_preprocess_raises_error_when_missing_dest_difference_var(
    df_with_cumulated: pd.DataFrame, preprocessed_df_with_cumulated: pd.DataFrame
) -> None:
    preprocessed_df_with_cumulated.loc[2, "value_difference"] = np.nan

    processor = InterRecordCumulatedDifferenceProcessor(
        id_variable="id",
        target_variable="value",
        new_first_variable_name="first_value",
        new_difference_variable_name="value_difference",
    )

    with pytest.raises(
        ValueError,
        match="Expected no missing values for `new_difference_variable_name`",
    ):
        processor.postprocess(df_with_cumulated, preprocessed_df_with_cumulated)


def test_preprocess_raises_error_when_wrong_id_var(
    df_with_cumulated: pd.DataFrame,
) -> None:
    processor = InterRecordCumulatedDifferenceProcessor(
        id_variable="wrong_id",
        target_variable="value",
        new_first_variable_name="first_value",
        new_difference_variable_name="value_difference",
    )

    with pytest.raises(ValueError, match="Expected a valid `id_variable`"):
        processor.preprocess(df_with_cumulated)


def test_preprocess_raises_error_when_wrong_target_var(
    df_with_cumulated: pd.DataFrame,
) -> None:
    processor = InterRecordCumulatedDifferenceProcessor(
        id_variable="id",
        target_variable="wrong_value",
        new_first_variable_name="first_value",
        new_difference_variable_name="value_difference",
    )

    with pytest.raises(ValueError, match="Expected a valid `target_variable`"):
        processor.preprocess(df_with_cumulated)


def test_postprocess_raises_error_when_wrong_first_var(
    df_with_cumulated: pd.DataFrame, preprocessed_df_with_cumulated: pd.DataFrame
) -> None:
    processor = InterRecordCumulatedDifferenceProcessor(
        id_variable="id",
        target_variable="value",
        new_first_variable_name="wrong_first_value",
        new_difference_variable_name="value_difference",
    )

    with pytest.raises(ValueError, match="Expected a valid `new_first_variable_name`"):
        processor.postprocess(df_with_cumulated, preprocessed_df_with_cumulated)


def test_postprocess_raises_error_when_wrong_difference_var(
    df_with_cumulated: pd.DataFrame, preprocessed_df_with_cumulated: pd.DataFrame
) -> None:
    processor = InterRecordCumulatedDifferenceProcessor(
        id_variable="id",
        target_variable="value",
        new_first_variable_name="first_value",
        new_difference_variable_name="wrong_value_difference",
    )

    with pytest.raises(
        ValueError, match="Expected a valid `new_difference_variable_name`"
    ):
        processor.postprocess(df_with_cumulated, preprocessed_df_with_cumulated)


def test_postprocess_raises_error_with_keep_order_and_different_indices(
    df_with_cumulated: pd.DataFrame,
) -> None:
    processor = InterRecordCumulatedDifferenceProcessor(
        id_variable="id",
        target_variable="value",
        new_first_variable_name="first_value",
        new_difference_variable_name="value_difference",
        keep_record_order=True,
    )

    processed_df = processor.preprocess(df_with_cumulated)
    # make indices of processed_df different than those of df_with_cumulated
    processed_df.index = range(10, 10 + len(processed_df), 1)

    with pytest.raises(
        ValueError,
        match="Expected `keep_record_order` to be `True` only if",
    ):
        processor.postprocess(df_with_cumulated, processed_df)



# This file has been generated - DO NOT MODIFY
# API Version : 10.2.0


import logging
import shutil
import tempfile
import warnings
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, TypeVar, Union
from uuid import UUID

import numpy as np
import pandas as pd

from avatars.api_autogenerated import Auth as _Auth
from avatars.api_autogenerated import Compatibility as _Compatibility
from avatars.api_autogenerated import Datasets as _Datasets
from avatars.api_autogenerated import Health as _Health
from avatars.api_autogenerated import Jobs as _Jobs
from avatars.api_autogenerated import Metrics as _Metrics
from avatars.api_autogenerated import Reports as _Reports
from avatars.api_autogenerated import Stats as _Stats
from avatars.api_autogenerated import Users as _Users
from avatars.arrow_utils import (
    ArrowDatasetBuilder,
    DataSourceItem,
    DataSourceItems,
    FileLike,
    is_file_like,
)
from avatars.client import ApiClient
from avatars.constants import DEFAULT_TIMEOUT
from avatars.exceptions import Timeout
from avatars.models import AdviceJob  # noqa: F401
from avatars.models import AdviceJobCreate  # noqa: F401
from avatars.models import AvatarizationJob  # noqa: F401
from avatars.models import AvatarizationJobCreate  # noqa: F401
from avatars.models import AvatarizationMultiTableJob  # noqa: F401
from avatars.models import AvatarizationMultiTableJobCreate  # noqa: F401
from avatars.models import AvatarizationWithTimeSeriesJob  # noqa: F401
from avatars.models import AvatarizationWithTimeSeriesJobCreate  # noqa: F401
from avatars.models import ClusterStats  # noqa: F401
from avatars.models import CompatibilityResponse  # noqa: F401
from avatars.models import Contributions  # noqa: F401
from avatars.models import CreateDataset  # noqa: F401
from avatars.models import CreateUser  # noqa: F401
from avatars.models import Dataset  # noqa: F401
from avatars.models import ExplainedVariance  # noqa: F401
from avatars.models import ForgottenPasswordRequest  # noqa: F401
from avatars.models import GenericJob  # noqa: F401
from avatars.models import Login  # noqa: F401
from avatars.models import LoginResponse  # noqa: F401
from avatars.models import PatchDataset  # noqa: F401
from avatars.models import PrivacyMetricsGeolocationJob  # noqa: F401
from avatars.models import PrivacyMetricsGeolocationJobCreate  # noqa: F401
from avatars.models import PrivacyMetricsJob  # noqa: F401
from avatars.models import PrivacyMetricsJobCreate  # noqa: F401
from avatars.models import PrivacyMetricsMultiTableJob  # noqa: F401
from avatars.models import PrivacyMetricsMultiTableJobCreate  # noqa: F401
from avatars.models import PrivacyMetricsWithTimeSeriesJob  # noqa: F401
from avatars.models import PrivacyMetricsWithTimeSeriesJobCreate  # noqa: F401
from avatars.models import Projections  # noqa: F401
from avatars.models import Report  # noqa: F401
from avatars.models import ReportCreate  # noqa: F401
from avatars.models import ReportFromDataCreate  # noqa: F401
from avatars.models import ReportGeolocationPrivacyCreate  # noqa: F401
from avatars.models import ResetPasswordRequest  # noqa: F401
from avatars.models import SignalMetricsJob  # noqa: F401
from avatars.models import SignalMetricsJobCreate  # noqa: F401
from avatars.models import SignalMetricsWithTimeSeriesJob  # noqa: F401
from avatars.models import SignalMetricsWithTimeSeriesJobCreate  # noqa: F401
from avatars.models import User  # noqa: F401
from avatars.models import (
    AvatarizationPipelineCreate,
    AvatarizationPipelineResult,
    ColumnDetail,
    ColumnType,
    FileType,
    JobStatus,
    PrivacyMetricsParameters,
    SignalMetricsParameters,
)

logger = logging.getLogger(__name__)

DEFAULT_RETRY_TIMEOUT = 60
MAX_ROWS_PER_FILE = 1_000_000
MAX_BYTES_PER_FILE = 100 * 1024 * 1024  # 100 MB

PARQUET_MAGIC_BYTES = b"PAR1"

T = TypeVar("T")


def to_column_type(s: str) -> ColumnType:
    if "float" in s:
        return ColumnType.float
    if "int" in s:
        return ColumnType.int
    if "bool" in s:
        return ColumnType.bool
    if "datetime" in s:
        return ColumnType.datetime
    if "object" in s or s == "category" or s == "str":
        return ColumnType.category
    raise TypeError(f"Unknown column type: '{s}'")


def from_column_type(s: ColumnType) -> str:
    # We don't want to use 'category' here, as we store it as 'object' in the API
    if s == ColumnType.category:
        return "object"
    else:
        return str(s.value)

    raise TypeError(f"Unknown column type: '{s}'")


class Auth:
    def __init__(self, client: ApiClient) -> None:
        self.client = client

    def login(
        self,
        request: Login,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> LoginResponse:
        """Login the user."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            request,
        ]

        return _Auth(self.client).login(*args, **kwargs)

    def refresh(
        self,
        token: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> LoginResponse:

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            token,
        ]

        return _Auth(self.client).refresh(*args, **kwargs)

    def forgotten_password(
        self,
        request: ForgottenPasswordRequest,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Any:

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            request,
        ]

        return _Auth(self.client).forgotten_password(*args, **kwargs)

    def reset_password(
        self,
        request: ResetPasswordRequest,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Any:

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            request,
        ]

        return _Auth(self.client).reset_password(*args, **kwargs)


class Compatibility:
    def __init__(self, client: ApiClient) -> None:
        self.client = client

    def is_client_compatible(
        self,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> CompatibilityResponse:
        """Verify if the client is compatible with the API."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = []

        return _Compatibility(self.client).is_client_compatible(*args, **kwargs)


class Datasets:
    def __init__(self, client: ApiClient) -> None:
        self.client = client

    def create_dataset_from_stream(
        self,
        request: Optional[FileLike] = None,
        name: Optional[str] = None,
        source: Optional[
            DataSourceItem
        ] = None,  # optional because we still have to support the old way # TODO: Remove once deprecated
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Dataset:
        """Create a dataset by streaming chunks of the dataset.

        DEPRECATED: Please use create_dataset instead.
        """

        warnings.warn(
            DeprecationWarning(
                "create_dataset_from_stream is deprecated. Use create_dataset instead."
            )
        )

        _source: Optional[Union[str, FileLike]] = request or source
        return self.create_dataset(source=_source, name=name, timeout=timeout)

    def create_dataset(
        self,
        request: Optional[FileLike] = None,  # TODO: Remove once deprecated
        name: Optional[str] = None,
        source: Optional[
            DataSourceItems
        ] = None,  # optional because we still have to support the old way
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Dataset:
        """Create a dataset from file upload."""

        if request:
            warnings.warn(
                DeprecationWarning("request is deprecated. Use file instead.")
            )

        if request is not None and source is not None:
            raise ValueError("You cannot pass both request and source.")

        if request is None and source is None:
            raise ValueError("You need to pass in a source.")

        _source = request or source

        if _source is None:
            raise ValueError("You need to pass in a source.")

        ds, inferred_type = ArrowDatasetBuilder().to_dataset(_source)

        filetype: Optional[FileType] = (
            FileType[inferred_type] if inferred_type else None
        )

        kwargs = {
            "method": "post",
            "url": "/datasets/stream",
            "timeout": timeout,
            "dataset": ds,
            "params": dict(name=name, filetype=filetype),
        }

        result = self.client.request(**kwargs)

        return Dataset(**result)

    def download_dataset_as_stream(
        self,
        id: str,
        destination: Optional[Union[str, FileLike]] = None,
        *,
        filetype: Optional[FileType] = None,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Any:
        """Download a dataset by streaming chunks of it.

        DEPRECATED: Please use download_dataset instead.
        """
        warnings.warn(
            DeprecationWarning(
                "download_dataset_as_stream is deprecated. Use download_dataset instead."
            )
        )

        # Ignoring return type because download_dataset's logic makes sure
        # that the return type will be BytesIO
        # No point in going through the hassle of using typing.overload for something
        # that will be removed soon
        return self.download_dataset(
            id,
            destination=destination,
            timeout=timeout,
            filetype=filetype,
            from_download_as_stream=True,
        )

    def download_dataset(
        self,
        id: str,
        destination: Optional[Union[str, FileLike]] = None,
        filetype: Optional[FileType] = None,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
        from_download_as_stream: bool = False,  # TODO: Remove once deprecated
    ) -> Any:
        """Download a dataset."""
        # Download dataset metadata
        dataset_info = self.client.datasets.get_dataset(id, timeout=timeout)

        if destination is None:
            warnings.warn(
                DeprecationWarning(
                    "Please specify the destination argument. The return type will change in the future."
                )
            )

            if not from_download_as_stream:
                # Download as CSV was the old return value when using download_dataset
                filetype = filetype or FileType.csv
            else:
                # Download as the filetype of the dataset on the server was the old
                # return value when using download_dataset_as_stream
                filetype = filetype or dataset_info.filetype
        else:
            if not (isinstance(destination, str) or is_file_like(destination)):
                raise TypeError(
                    f"Expected destination to be a string or a buffer, got {type(destination)} instead"
                )

        return self.client.request(
            method="get",
            url=f"/datasets/{id}/download/stream",
            params={"filetype": filetype},
            should_stream=True,
            want_content=destination is None and not from_download_as_stream,
            timeout=timeout,
            destination=destination,
        )

    def find_all_datasets_by_user(
        self,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> List[Dataset]:
        """List all datasets of the current_user."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = []

        return _Datasets(self.client).find_all_datasets_by_user(*args, **kwargs)

    def get_dataset(
        self,
        id: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Dataset:
        """Get a dataset."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            id,
        ]

        return _Datasets(self.client).get_dataset(*args, **kwargs)

    def patch_dataset(
        self,
        request: PatchDataset,
        id: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Dataset:
        """Modify a dataset."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            request,
            id,
        ]

        return _Datasets(self.client).patch_dataset(*args, **kwargs)

    def analyze_dataset(
        self,
        id: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Dataset:
        """Start the analysis of a dataset."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            id,
        ]

        return _Datasets(self.client).analyze_dataset(*args, **kwargs)

    def get_dataset_correlations(
        self,
        id: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Any:
        """Get a dataset's correlations."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            id,
        ]

        return _Datasets(self.client).get_dataset_correlations(*args, **kwargs)


class Health:
    def __init__(self, client: ApiClient) -> None:
        self.client = client

    def get_root(
        self,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Any:
        """Verify server health."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = []

        return _Health(self.client).get_root(*args, **kwargs)

    def get_health(
        self,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Any:
        """Verify server health."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = []

        return _Health(self.client).get_health(*args, **kwargs)

    def get_health_db(
        self,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Any:
        """Verify connection to the db health."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = []

        return _Health(self.client).get_health_db(*args, **kwargs)


class Jobs:
    def __init__(self, client: ApiClient) -> None:
        self.client = client

    def find_all_jobs_by_user(
        self,
        nb_days: Optional[int] = None,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> List[GenericJob]:
        """Retrieve all jobs executed by the current user.

        Jobs are filtered by execution date, by default only the last 5 days are displayed,
        a parameter can be provided to go further back in time.
        """

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            nb_days,
        ]

        return _Jobs(self.client).find_all_jobs_by_user(*args, **kwargs)

    def create_full_avatarization_job(
        self,
        request: AvatarizationJobCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> AvatarizationJob:
        """Create an avatarization job, then calculate metrics."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            request,
        ]

        return _Jobs(self.client).create_full_avatarization_job(*args, **kwargs)

    def cancel_job(
        self,
        id: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> GenericJob:
        """Cancel any kind of job.

        If the job hadn't been started yet, revoke it.
        If the job is ongoing, gently kill it.
        If the job is done, do nothing.
        """

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            id,
        ]

        return _Jobs(self.client).cancel_job(*args, **kwargs)

    def create_avatarization_job(
        self,
        request: AvatarizationJobCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> AvatarizationJob:
        """Create an avatarization job."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            request,
        ]

        return _Jobs(self.client).create_avatarization_job(*args, **kwargs)

    def create_avatarization_with_time_series_job(
        self,
        request: AvatarizationWithTimeSeriesJobCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> AvatarizationWithTimeSeriesJob:
        """Create an avatarization with time series job."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            request,
        ]

        return _Jobs(self.client).create_avatarization_with_time_series_job(
            *args, **kwargs
        )

    def create_avatarization_multi_table_job(
        self,
        request: AvatarizationMultiTableJobCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> AvatarizationMultiTableJob:
        """Create an avatarization for relational data."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            request,
        ]

        return _Jobs(self.client).create_avatarization_multi_table_job(*args, **kwargs)

    def create_signal_metrics_job(
        self,
        request: SignalMetricsJobCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> SignalMetricsJob:
        """Create a signal metrics job."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            request,
        ]

        return _Jobs(self.client).create_signal_metrics_job(*args, **kwargs)

    def create_privacy_metrics_job(
        self,
        request: PrivacyMetricsJobCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> PrivacyMetricsJob:
        """Create a privacy metrics job."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            request,
        ]

        return _Jobs(self.client).create_privacy_metrics_job(*args, **kwargs)

    def create_privacy_metrics_time_series_job(
        self,
        request: PrivacyMetricsWithTimeSeriesJobCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> PrivacyMetricsWithTimeSeriesJob:
        """Create a privacy metrics with time series job."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            request,
        ]

        return _Jobs(self.client).create_privacy_metrics_time_series_job(
            *args, **kwargs
        )

    def create_signal_metrics_time_series_job(
        self,
        request: SignalMetricsWithTimeSeriesJobCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> SignalMetricsWithTimeSeriesJob:
        """Create a signal metrics with time series job."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            request,
        ]

        return _Jobs(self.client).create_signal_metrics_time_series_job(*args, **kwargs)

    def create_privacy_metrics_multi_table_job(
        self,
        request: PrivacyMetricsMultiTableJobCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> PrivacyMetricsMultiTableJob:
        """Create a privacy metrics job."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            request,
        ]

        return _Jobs(self.client).create_privacy_metrics_multi_table_job(
            *args, **kwargs
        )

    def create_privacy_metrics_geolocation_job(
        self,
        request: PrivacyMetricsGeolocationJobCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> PrivacyMetricsGeolocationJob:
        """Create a geolocation privacy metrics job."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            request,
        ]

        return _Jobs(self.client).create_privacy_metrics_geolocation_job(
            *args, **kwargs
        )

    def get_privacy_metrics_geolocation_job(
        self,
        id: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> PrivacyMetricsGeolocationJob:
        """Get a geolocation privacy metrics job."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            id,
        ]

        return _Jobs(self.client).get_privacy_metrics_geolocation_job(*args, **kwargs)

    def get_avatarization_job(
        self,
        id: str,
        *,
        per_request_timeout: Optional[int] = DEFAULT_TIMEOUT,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> AvatarizationJob:
        """Get an avatarization job."""

        kwargs: Dict[str, Any] = {
            "per_request_timeout": per_request_timeout,
            "timeout": timeout,
        }

        args: List[Any] = [
            id,
        ]

        return _Jobs(self.client).get_avatarization_job(*args, **kwargs)

    def get_avatarization_time_series_job(
        self,
        id: str,
        *,
        per_request_timeout: Optional[int] = DEFAULT_TIMEOUT,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> AvatarizationWithTimeSeriesJob:
        """Get an avatarization time series job."""

        kwargs: Dict[str, Any] = {
            "per_request_timeout": per_request_timeout,
            "timeout": timeout,
        }

        args: List[Any] = [
            id,
        ]

        return _Jobs(self.client).get_avatarization_time_series_job(*args, **kwargs)

    def get_avatarization_multi_table_job(
        self,
        id: str,
        *,
        per_request_timeout: Optional[int] = DEFAULT_TIMEOUT,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> AvatarizationMultiTableJob:
        """Get a multi table avatarization job."""

        kwargs: Dict[str, Any] = {
            "per_request_timeout": per_request_timeout,
            "timeout": timeout,
        }

        args: List[Any] = [
            id,
        ]

        return _Jobs(self.client).get_avatarization_multi_table_job(*args, **kwargs)

    def get_signal_metrics(
        self,
        id: str,
        *,
        per_request_timeout: Optional[int] = DEFAULT_TIMEOUT,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> SignalMetricsJob:
        """Get a signal metrics job."""

        kwargs: Dict[str, Any] = {
            "per_request_timeout": per_request_timeout,
            "timeout": timeout,
        }

        args: List[Any] = [
            id,
        ]

        return _Jobs(self.client).get_signal_metrics(*args, **kwargs)

    def get_signal_metrics_time_series_job(
        self,
        id: str,
        *,
        per_request_timeout: Optional[int] = DEFAULT_TIMEOUT,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> SignalMetricsWithTimeSeriesJob:
        """Get a signal metrics time series job."""

        kwargs: Dict[str, Any] = {
            "per_request_timeout": per_request_timeout,
            "timeout": timeout,
        }

        args: List[Any] = [
            id,
        ]

        return _Jobs(self.client).get_signal_metrics_time_series_job(*args, **kwargs)

    def get_privacy_metrics(
        self,
        id: str,
        *,
        per_request_timeout: Optional[int] = DEFAULT_TIMEOUT,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> PrivacyMetricsJob:
        """Get a privacy metrics job."""

        kwargs: Dict[str, Any] = {
            "per_request_timeout": per_request_timeout,
            "timeout": timeout,
        }

        args: List[Any] = [
            id,
        ]

        return _Jobs(self.client).get_privacy_metrics(*args, **kwargs)

    def get_privacy_metrics_time_series_job(
        self,
        id: str,
        *,
        per_request_timeout: Optional[int] = DEFAULT_TIMEOUT,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> PrivacyMetricsWithTimeSeriesJob:
        """Get a privacy metrics time series job."""

        kwargs: Dict[str, Any] = {
            "per_request_timeout": per_request_timeout,
            "timeout": timeout,
        }

        args: List[Any] = [
            id,
        ]

        return _Jobs(self.client).get_privacy_metrics_time_series_job(*args, **kwargs)

    def get_privacy_metrics_multi_table_job(
        self,
        id: str,
        *,
        per_request_timeout: Optional[int] = DEFAULT_TIMEOUT,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> PrivacyMetricsMultiTableJob:
        """Get a privacy metrics multi table job."""

        kwargs: Dict[str, Any] = {
            "per_request_timeout": per_request_timeout,
            "timeout": timeout,
        }

        args: List[Any] = [
            id,
        ]

        return _Jobs(self.client).get_privacy_metrics_multi_table_job(*args, **kwargs)

    def create_advice(
        self,
        request: AdviceJobCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> AdviceJob:
        """Create advice on anonymization parameters."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            request,
        ]

        return _Jobs(self.client).create_advice(*args, **kwargs)

    def get_advice(
        self,
        id: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> AdviceJob:
        """Get advice result."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            id,
        ]

        return _Jobs(self.client).get_advice(*args, **kwargs)


class Metrics:
    def __init__(self, client: ApiClient) -> None:
        self.client = client

    def get_job_projections(
        self,
        job_id: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Projections:
        """Get the projections of records and avatars in 3D.

        See https://saiph.readthedocs.io/en/latest/ for more information.

        Arguments
        ---------
            job_id:
                avatarization or privacy job id used to fit the model
        """

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            job_id,
        ]

        return _Metrics(self.client).get_job_projections(*args, **kwargs)

    def get_variable_contributions(
        self,
        job_id: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Contributions:
        """Get the contributions of the dataset variables within the fitted space.

        See https://saiph.readthedocs.io/en/latest for more information.

        Arguments
        ---------
            job_id:
                avatarization or privacy job id used to fit the model
        """

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            job_id,
        ]

        return _Metrics(self.client).get_variable_contributions(*args, **kwargs)

    def get_explained_variance(
        self,
        job_id: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> ExplainedVariance:
        """Get the explained variance of records.

        See https://saiph.readthedocs.io/en/latest/ for more information.

        Arguments
        ---------
            job_id:
                avatarization or privacy job id used to fit the model
        """

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            job_id,
        ]

        return _Metrics(self.client).get_explained_variance(*args, **kwargs)


class Reports:
    def __init__(self, client: ApiClient) -> None:
        self.client = client

    def create_report(
        self,
        request: ReportCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Report:
        """Create an anonymization report."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            request,
        ]

        return _Reports(self.client).create_report(*args, **kwargs)

    def get_report(
        self,
        id: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Report:

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            id,
        ]

        return _Reports(self.client).get_report(*args, **kwargs)

    def download_report(
        self,
        id: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Any:
        """Download a report."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            id,
        ]

        return _Reports(self.client).download_report(*args, **kwargs)

    def create_report_from_data(
        self,
        request: ReportFromDataCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Report:
        """Create an anonymization report without avatarization job."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            request,
        ]

        return _Reports(self.client).create_report_from_data(*args, **kwargs)

    def create_geolocation_privacy_report(
        self,
        request: ReportGeolocationPrivacyCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> Report:
        """Create an anonymization report without avatarization job."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            request,
        ]

        return _Reports(self.client).create_geolocation_privacy_report(*args, **kwargs)


class Stats:
    def __init__(self, client: ApiClient) -> None:
        self.client = client

    def get_cluster_stats(
        self,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> ClusterStats:
        """Get insights into the cluster's usage."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = []

        return _Stats(self.client).get_cluster_stats(*args, **kwargs)


class Users:
    def __init__(self, client: ApiClient) -> None:
        self.client = client

    def find_users(
        self,
        email: Optional[str] = None,
        username: Optional[str] = None,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> List[User]:
        """Get users, optionally filtering them by username or email.

        This endpoint is protected with rate limiting.
        """

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            email,
            username,
        ]

        return _Users(self.client).find_users(*args, **kwargs)

    def create_user(
        self,
        request: CreateUser,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> User:
        """Create a user.

        This endpoint is protected with rate limiting.
        """

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            request,
        ]

        return _Users(self.client).create_user(*args, **kwargs)

    def get_me(
        self,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> User:
        """Get my own user."""

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = []

        return _Users(self.client).get_me(*args, **kwargs)

    def get_user(
        self,
        id: str,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> User:
        """Get a user by id.

        This endpoint is protected with rate limiting.
        """

        kwargs: Dict[str, Any] = {
            "timeout": timeout,
        }

        args: List[Any] = [
            id,
        ]

        return _Users(self.client).get_user(*args, **kwargs)


class PandasIntegration:
    def __init__(self, client: ApiClient) -> None:
        self.client = client

    def upload_dataframe(
        self,
        request: "pd.DataFrame",
        name: Optional[str] = None,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
        identifier_variables: List[str] = [],
        **kwargs: Dict[str, Any],  # to collect should_stream
    ) -> Dataset:

        # Reset the index of the dataframe
        request = request.reset_index(drop=True)

        if "should_stream" in kwargs:
            warnings.warn(
                "The `should_stream` parameter is deprecated and will be removed in a future version. "
                "All uploads are now streamed by default.",
                DeprecationWarning,
            )

        if not set(identifier_variables).issubset(set(request.columns)):
            missing_columns = set(identifier_variables) - set(request.columns)
            raise ValueError(
                "Expected all identifier variables to be present in the dataframe."
                f" Got {missing_columns}"
            )

        for col in request.columns:
            if pd.api.types.infer_dtype(request[col], skipna=True) in (
                "mixed-integer",
                "mixed",
            ):
                raise ValueError(
                    f"Expected column '{col}' to have either str or numeric values."
                    " Consider harmonizing columns prior to upload."
                )

        df_types = request.dtypes

        dataset = self.client.datasets.create_dataset(
            source=request,
            name=name,
            timeout=timeout,
        )

        columns = []
        for index, dtype in zip(df_types.index, df_types):
            column_detail = ColumnDetail(
                type=to_column_type(str(dtype)),
                label=index,
                is_identifier=index in identifier_variables,
            )
            columns.append(column_detail)

        dataset = self.client.datasets.patch_dataset(
            id=str(dataset.id),
            request=PatchDataset(columns=columns),
        )
        return dataset

    def download_dataframe(
        self,
        id: str,
        *,
        filetype: Optional[FileType] = None,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
        **kwargs: Any,  # to collect should_stream
    ) -> pd.DataFrame:
        if "should_stream" in kwargs:
            warnings.warn(
                "The `should_stream` parameter is deprecated and will be removed in a future version. "
                "All uploads are now streamed by default.",
                DeprecationWarning,
            )

        dataset_info = self.client.datasets.get_dataset(id, timeout=timeout)

        _filetype = filetype or dataset_info.filetype
        with tempfile.TemporaryDirectory() as download_dir:
            path = Path(download_dir) / "file"

            try:
                Datasets(client=self.client).download_dataset(
                    id, str(path), timeout=timeout, filetype=_filetype
                )
                if _filetype is FileType.parquet:
                    df = pd.read_parquet(path, engine="pyarrow")
                elif _filetype is FileType.csv:
                    df = pd.read_csv(path)
                else:
                    raise ValueError(f"Unsupported filetype: {_filetype}")
            except Exception as e:
                # Save file somewhere to allow for investigation
                tdir = tempfile.mkdtemp()
                tfile = Path(tdir) / f"dataframe-{id}.fail"
                shutil.copyfile(path, tfile)
                warnings.warn(f"download failed: file is here: {str(tfile)}")
                raise e

        # We apply datetime columns separately as 'datetime' is not a valid pandas dtype
        dtypes = {c.label: from_column_type(c.type) for c in dataset_info.columns or {}}
        datetime_columns = [
            label for label, type in dtypes.items() if type == ColumnType.datetime.value
        ]

        # Remove datetime columns
        for label in list(dtypes.keys()):
            if label in datetime_columns:
                dtypes.pop(label, None)

        for name, dtype in dtypes.items():
            df[name] = df[name].astype(dtype)

        df[datetime_columns] = df[datetime_columns].astype("datetime64[ns]")

        return df


class Pipelines:
    def __init__(self, client: ApiClient) -> None:
        self.client = client

    def avatarization_pipeline_with_processors(
        self,
        request: AvatarizationPipelineCreate,
        *,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
        per_request_timeout: Optional[int] = DEFAULT_TIMEOUT,
    ) -> AvatarizationPipelineResult:
        """Create a Pipeline of avatarization with processors."""

        # Upload the dataset if needed
        # User can either specify the dataset he already uploaded or let this pipeline upload it
        df = request.df.copy()
        original_dataset_id = (
            request.avatarization_job_create.parameters.dataset_id
            or self.client.pandas_integration.upload_dataframe(df, timeout=timeout).id
        )

        # Pre process the dataframe and upload it
        processors = request.processors
        for p in processors:
            df = p.preprocess(df)
        dataset = self.client.pandas_integration.upload_dataframe(
            df, timeout=per_request_timeout
        )

        # Avatarize the uploaded dataframe
        request.avatarization_job_create.parameters.dataset_id = dataset.id
        avatarization_job = self.client.jobs.create_avatarization_job(
            request.avatarization_job_create
        )
        logger.info(f"launching avatarization job with id={avatarization_job.id}")

        avatarization_job = self.client.jobs.get_avatarization_job(
            str(avatarization_job.id),
            timeout=timeout,
            per_request_timeout=per_request_timeout,
        )
        if avatarization_job.status == JobStatus.failure:
            raise Exception(
                f"Got error during the avatarization job: {avatarization_job.error_message}"
            )

        if (
            avatarization_job.status == JobStatus.pending
            or not avatarization_job.result
        ):
            raise Timeout(
                f"The avatarization job '{avatarization_job.id}' timed out."
                ""
                """Try increasing the timeout with the `timeout` parameter."""
            )

        # Download the dataframe, postprocess it and upload the new dataframe
        sensitive_unshuffled_avatars = (
            self.client.pandas_integration.download_dataframe(
                str(avatarization_job.result.sensitive_unshuffled_avatars_datasets.id),
                timeout=timeout,
            )
        )
        for p in reversed(processors):
            sensitive_unshuffled_avatars = p.postprocess(
                request.df, sensitive_unshuffled_avatars
            )
        unshuffled_dataset = self.client.pandas_integration.upload_dataframe(
            sensitive_unshuffled_avatars, timeout=timeout
        )

        # Calculate privacy metrics on the post processed dataset vs the original one
        privacy_job = self.client.jobs.create_privacy_metrics_job(
            PrivacyMetricsJobCreate(
                parameters=PrivacyMetricsParameters(
                    original_id=original_dataset_id,
                    unshuffled_avatars_id=unshuffled_dataset.id,
                )
            ),
            timeout=per_request_timeout,
        )
        logger.info(f"launching privacy metrics job with id={privacy_job.id}")

        # Calculate signal metrics
        signal_job = self.client.jobs.create_signal_metrics_job(
            SignalMetricsJobCreate(
                parameters=SignalMetricsParameters(
                    original_id=original_dataset_id, avatars_id=unshuffled_dataset.id
                )
            ),
            timeout=per_request_timeout,
        )
        logger.info(f"launching signal metrics job with id={signal_job.id}")

        # Get the job results
        signal_job = self.client.jobs.get_signal_metrics(
            str(signal_job.id), timeout=timeout, per_request_timeout=per_request_timeout
        )
        privacy_job = self.client.jobs.get_privacy_metrics(
            str(privacy_job.id),
            timeout=timeout,
            per_request_timeout=per_request_timeout,
        )
        if signal_job.status == JobStatus.failure:
            raise Exception(
                f"Got error during the signal metrics job: {signal_job.error_message}"
            )
        if privacy_job.status == JobStatus.failure:
            raise Exception(
                f"Got error during the privacy metrics job: {privacy_job.error_message}"
            )

        if signal_job.status == JobStatus.pending or not signal_job.result:
            raise Timeout(
                f"The signal metrics job '{signal_job.id}' timed out."
                ""
                """Try increasing the timeout with the `timeout` parameter."""
            )

        if privacy_job.status == JobStatus.pending or not privacy_job.result:
            raise Timeout(
                f"The privacy metrics job '{privacy_job.id}' timed out."
                ""
                """Try increasing the timeout with the `timeout` parameter."""
            )

        # Shuffle sensitive_unshuffled_avatars for security reasons
        random_gen = np.random.default_rng()
        map = random_gen.permutation(sensitive_unshuffled_avatars.index.values).tolist()
        post_processed_avatars = sensitive_unshuffled_avatars.iloc[map].reset_index(
            drop=True
        )

        return AvatarizationPipelineResult(
            privacy_metrics=privacy_job.result,
            signal_metrics=signal_job.result,
            post_processed_avatars=post_processed_avatars,
            avatarization_job_id=avatarization_job.id,
            signal_job_id=signal_job.id,
            privacy_job_id=privacy_job.id,
        )


def upload_batch_and_get_order(
    client: ApiClient,
    training: pd.DataFrame,
    splits: List[pd.DataFrame],
    timeout: int = DEFAULT_TIMEOUT,
) -> Tuple[UUID, List[UUID], Dict[UUID, pd.Index]]:
    """Upload batches to the server
    Arguments
    ---------
        client:
            Api client
        training:
            Dataframe used to train the anonymization model. This dataframe should contain all modalities of the categorical variables.
        splits:
            All other batches
    Returns
    -------
        training_dataset_id:
            The dataset id of the training dataset
        datasets_split_ids:
            The dataset id of all other batches
        batch_mapping:
            The index mapping for each dataset batch
    """
    training_dataset = client.pandas_integration.upload_dataframe(
        training, timeout=timeout
    )

    datasets_split_ids = [
        client.pandas_integration.upload_dataframe(split, timeout=timeout).id
        for split in splits
    ]
    batch_mapping: Dict[UUID, pd.Index] = {training_dataset.id: training.index}
    for dataset, dataframe in zip(datasets_split_ids, splits):
        batch_mapping[dataset] = dataframe.index

    return training_dataset.id, datasets_split_ids, batch_mapping


# This file has been generated - DO NOT MODIFY



import numpy as np
import pandas as pd

from avatars.lib.continuous_threshold import get_continuous_under_threshold


class ToCategoricalProcessor:
    """Processor to model selected numeric variables as categorical variables.

    Arguments
    ---------
        to_categorical_threshold:
            threshold of the number of distinct value to consider
            a continuous variable as categorical.

    Keyword Arguments
    -----------------
        keep_continuous:
            if `True`, continuous variables will be kept and
        suffixed with `continuous_suffix`.
        continuous_suffix:
            suffix for the continuous variable created during preprocess.
        category:
            if `keep_continuous=True`, name of the new category, needed for some specific
            avatarization cases with the use of group_modalities processor

    Examples
    --------
    With `keep_continuous=False` it only convert the variable to object.
    By this you ensure to keep all values during the avatarization.

    >>> df = pd.DataFrame(
    ...    {
    ...        "variable_1": [1, 7, 7, 1],
    ...        "variable_2": [1, 2, 7, 1]
    ...        }
    ...    )
    >>> processor = ToCategoricalProcessor(to_categorical_threshold = 2)
    >>> processor.preprocess(df).dtypes
    variable_1    object
    variable_2     int64
    dtype: object
    >>> avatar = pd.DataFrame(
    ...    {
    ...        "variable_1": [2, 1, 4, 1],
    ...        "variable_2": [2, 1, 4, 1]
    ...        }
    ...    )
    >>> avatar["variable_1"] = avatar["variable_1"].astype('object')
    >>> avatar.dtypes
    variable_1    object
    variable_2     int64
    dtype: object
    >>> processor.postprocess(df, avatar).dtypes
    variable_1    int64
    variable_2    int64
    dtype: object

    With `keep_continuous=True`, you duplicate the variable and keep it as continuous.
    This can be useful for other uses.

    >>> df = pd.DataFrame(
    ...    {
    ...        "variable_1": [1, 7, 7, 1],
    ...        "variable_2": [1, 2, 7, 1]
    ...        }
    ...    )
    >>> processor = ToCategoricalProcessor(to_categorical_threshold=2, keep_continuous=True)
    >>> processor.preprocess(df).dtypes
    variable_1          object
    variable_2           int64
    variable_1__cont     int64
    dtype: object
    """

    def __init__(
        self,
        to_categorical_threshold: int,
        *,
        keep_continuous: bool = False,
        continuous_suffix: str = "__cont",
        category: str = "other",
    ):
        self.to_categorical_threshold = to_categorical_threshold
        self.keep_continuous = keep_continuous
        self.continuous_suffix = continuous_suffix
        self.category = category

    def preprocess(self, df: pd.DataFrame) -> pd.DataFrame:
        """Transform numeric variables into categorical variables.

        Arguments
        ---------
            df: dataframe to transform

        Returns
        -------
            DataFrame: transformed dataframe
        """
        df = df.copy()
        self.variables = get_continuous_under_threshold(
            df, threshold=self.to_categorical_threshold
        )

        # Perform the transformation by creating a new column cont_suffix where
        # NAs are kept as NAs.
        if self.keep_continuous:
            self.continuous_variables = [
                x + self.continuous_suffix for x in self.variables
            ]
            df[self.continuous_variables] = df[self.variables]

        df[self.variables] = df[self.variables].astype("object")
        return df

    def postprocess(self, source: pd.DataFrame, dest: pd.DataFrame) -> pd.DataFrame:
        """Transform converted categorical variables back to numeric.

        Arguments
        ---------
            source: reference data frame
            dest: data frame to transform

        Returns
        -------
            DataFrame: transformed data frame
        """
        dest = dest.copy()

        # affect the continuous value to modality "category",
        # needed for some specific avatarization cases with the use of group_modalities processor
        if self.keep_continuous:
            dest[self.variables] = np.where(
                dest[self.variables] == self.category,
                dest[self.continuous_variables],
                dest[self.variables],
            )
            dest = dest.drop(columns=self.continuous_variables)

        # reassign the original data type
        dest[self.variables] = dest[self.variables].astype(
            source[self.variables].dtypes.to_dict()
        )

        return dest



from typing import Optional

import numpy as np
import pandas as pd


class InterRecordBoundedRangeDifferenceProcessor:
    """Processor to express two related bounded variables relative to previous records.

    This processor can be used only on data where there are several records for each individual.
    By this transformation, variables such as `var_a` and `var_b` whose values are cumulative over
    successive records `t` in the following way:
    var_a_t <= var_b_t <= var_a_t+1 <= var_b_t+1 <= var_a_t+2 ...

    will be expressed as:
    - a variable containing the first value of `var_a`.
    - a variable containing the difference from the previous record
    - a variable containing the range between the start and end variables

    Difference and range variables are expressed as proportion of possible change between the
    value and the bound (upper or lower). For example, for a variable whose value only spreads from
    10 (lower bound) to 100 (upper bound), if the previous records value is 60 and the new value
    is 30, the proportion will be calculated as (30 - 60) / (60 - 10) = -0.6

    This processor is not suitable for data where any of the variables passed as args contain
    missing values.

    Keyword Arguments
    -----------------
        id_variable:
            variable indicating which individual each row belongs to
        target_start_variable:
            variable representing the start of the range to transform
        target_end_variable:
            variable representing the end of the range to transform
        sort_by_variable:
            variable used to sort records for each id
        new_first_variable:
            name of the variable to be created to contain the first value of the target variable
        new_range_variable:
            name of the variable to be created to contain the range value
        new_difference_variable:
            name of the variable to be created to contain the difference value
        should_round_output:
            set to True to force post-processed values to be integer.

    Examples
    --------
    >>> df = pd.DataFrame(
    ...    {
    ...       'quantity_start': [30, 100, 80, 70, 40, 70],
    ...       'quantity_end': [10, 80, 70, 60, 30, 5],
    ...       'b': [4, 3, 0, 0, 2, 4],
    ...       'id': [1,1,1,2,2,2]
    ...    }
    ... )
    >>> processor = InterRecordBoundedRangeDifferenceProcessor(
    ...    id_variable='id',
    ...    target_start_variable='quantity_start',
    ...    target_end_variable='quantity_end',
    ...    new_first_variable_name='quantity_s_first_val',
    ...    new_difference_variable_name="quantity_diff_to_bound",
    ...    new_range_variable="quantity_range",
    ...    should_round_output=True
    ... )
    >>> preprocessed_df = processor.preprocess(df)
    >>> print(preprocessed_df)
       b  id  quantity_range  quantity_s_first_val  quantity_diff_to_bound
    0  4   1       -0.800000                    30                0.000000
    1  3   1       -0.210526                    30                1.000000
    2  0   1       -0.133333                    30                0.000000
    3  0   2       -0.153846                    70                0.000000
    4  2   2       -0.285714                    70               -0.363636
    5  4   2       -1.000000                    70                0.571429

    The postprocess allows you to transform some preprocessed data back into its original format.

    >>> processor.postprocess(df, preprocessed_df)
       quantity_start  quantity_end  b  id
    0              30            10  4   1
    1             100            80  3   1
    2              80            70  0   1
    3              70            60  0   2
    4              40            30  2   2
    5              70             5  4   2
    """

    def __init__(
        self,
        *,
        id_variable: str,
        target_start_variable: str,
        target_end_variable: str,
        new_first_variable_name: str,
        new_range_variable: str,
        new_difference_variable_name: str,
        sort_by_variable: Optional[str] = None,
        should_round_output: bool = True,
    ):
        self.id_variable = id_variable
        self.target_start_variable = target_start_variable
        self.target_end_variable = target_end_variable
        self.new_first_variable_name = new_first_variable_name
        self.new_range_variable = new_range_variable
        self.new_difference_variable_name = new_difference_variable_name
        self.sort_by_variable = sort_by_variable
        self.should_round_output = should_round_output

    def preprocess(self, df: pd.DataFrame) -> pd.DataFrame:
        variables_to_check = [
            self.id_variable,
            self.target_start_variable,
            self.target_end_variable,
        ]
        if self.sort_by_variable:
            variables_to_check.append(self.sort_by_variable)
        if len(set(variables_to_check).difference(df.columns.values)) > 0:
            msg = "Expected valid variable names for `id_variable`, `target_start_variable`, "
            msg += f"`target_end_variable` and `sort_by_variable`, got '{self.id_variable}',"
            msg += f" '{self.target_start_variable}', '{self.target_end_variable}' and "
            msg += f"'{self.sort_by_variable}' instead"
            raise ValueError(msg)
        if df[variables_to_check].isnull().values.any():
            msg = "Expected no missing values for `id_variable`, `target_start_variable`, "
            msg += "`target_end_variable` and `sort_by_variable`, "
            msg += "got columns with nulls instead"
            raise ValueError(msg)

        working = df.copy()

        # sort records for each individual
        if self.sort_by_variable:
            working = working.sort_values([self.id_variable, self.sort_by_variable])
        else:
            working = working.sort_values([self.id_variable])

        # determine lb and ub
        working["lb"] = min(
            min(working[self.target_start_variable]),
            min(working[self.target_end_variable]),
        )
        working["ub"] = max(
            max(working[self.target_start_variable]),
            max(working[self.target_end_variable]),
        )

        # compute relative range to ub or lb
        working["range_increase"] = (
            working[self.target_end_variable] >= working[self.target_start_variable]
        )
        working["range_decrease"] = (
            working[self.target_end_variable] < working[self.target_start_variable]
        )
        working["diff_to_lb"] = abs(working[self.target_start_variable] - working["lb"])
        working["diff_to_ub"] = abs(working[self.target_start_variable] - working["ub"])
        working["relative_range_to_lb"] = (
            -abs(
                working[self.target_end_variable] - working[self.target_start_variable]
            )
            / working["diff_to_lb"]
        )
        working["relative_range_to_lb"] = working["relative_range_to_lb"].replace(
            [np.inf, -np.inf, np.nan], 0
        )
        working["relative_range_to_ub"] = (
            abs(working[self.target_end_variable] - working[self.target_start_variable])
            / working["diff_to_ub"]
        )
        working["relative_range_to_ub"] = working["relative_range_to_ub"].replace(
            [np.inf, -np.inf, np.nan], 0
        )
        working[self.new_range_variable] = (
            working["range_decrease"] * working["relative_range_to_lb"]
            + working["range_increase"] * working["relative_range_to_ub"]
        )

        # compute first value
        working[self.new_first_variable_name] = working.groupby(self.id_variable)[
            self.target_start_variable
        ].transform("first")

        # compute difference from previous val as proportion to lb or ub
        working["previous_val"] = working.groupby(self.id_variable)[
            self.target_end_variable
        ].shift()
        working = working.reset_index(drop=False)
        working.loc[working["previous_val"].isnull(), "previous_val"] = working[
            self.target_start_variable
        ]  # for first record, set previous value as same value to avoid NaN
        working["increase"] = (
            working[self.target_start_variable] >= working["previous_val"]
        )
        working["decrease"] = (
            working[self.target_start_variable] < working["previous_val"]
        )
        working["diff_to_lb"] = abs(working["previous_val"] - working["lb"])
        working["diff_to_ub"] = abs(working["previous_val"] - working["ub"])
        working["relative_diff_to_lb"] = (
            -abs(working[self.target_start_variable] - working["previous_val"])
            / working["diff_to_lb"]
        )
        working["relative_diff_to_lb"] = working["relative_diff_to_lb"].replace(
            [np.inf, -np.inf, np.nan], 0
        )
        working["relative_diff_to_ub"] = (
            abs(working[self.target_start_variable] - working["previous_val"])
            / working["diff_to_ub"]
        )
        working["relative_diff_to_ub"] = working["relative_diff_to_ub"].replace(
            [np.inf, -np.inf, np.nan], 0
        )
        working[self.new_difference_variable_name] = (
            working["decrease"] * working["relative_diff_to_lb"]
            + working["increase"] * working["relative_diff_to_ub"]
        )

        # Remove tmp variables
        working = working.drop(
            columns=[
                "relative_diff_to_lb",
                "relative_diff_to_ub",
                "previous_val",
                "diff_to_lb",
                "diff_to_ub",
                "decrease",
                "increase",
                "lb",
                "ub",
                self.target_start_variable,
                self.target_end_variable,
            ]
        )
        working = working.drop(
            columns=[
                "range_increase",
                "range_decrease",
                "relative_range_to_lb",
                "relative_range_to_ub",
            ]
        )

        # Re-order data as it was in the input - required because of earlier sorting step
        working = working.set_index("index").sort_index()
        working.index.name = None

        return working

    def postprocess(self, source: pd.DataFrame, dest: pd.DataFrame) -> pd.DataFrame:
        working = dest.copy()
        vals_s = []
        vals_e = []

        working["original_order"] = range(len(working))
        working = working.sort_values(by=[self.id_variable])

        # determine lb and ub
        lb = min(
            min(source[self.target_start_variable]),
            min(source[self.target_end_variable]),
        )
        ub = max(
            max(source[self.target_start_variable]),
            max(source[self.target_end_variable]),
        )

        # identify whether the difference and range values have increased or decreased
        working["increase"] = working[self.new_difference_variable_name] >= 0
        working["decrease"] = ~working["increase"]
        working[self.new_difference_variable_name] = abs(
            working[self.new_difference_variable_name]
        )
        working["range_increase"] = working[self.new_range_variable] >= 0
        working["range_decrease"] = ~working["range_increase"]

        # Iteratively compute the values based on the values of previous records
        # NB: It is not clear how this could be vectorized
        for theid in working[self.id_variable].unique():
            previous_val = working[working[self.id_variable] == theid][
                self.new_first_variable_name
            ].values[0]
            for (
                increase,
                decrease,
                relative_diff_to_bound,
                range_diff,
                range_increase,
                range_decrease,
            ) in working.loc[
                working[self.id_variable] == theid,
                [
                    "increase",
                    "decrease",
                    self.new_difference_variable_name,
                    self.new_range_variable,
                    "range_increase",
                    "range_decrease",
                ],
            ].values:
                val_s = (
                    previous_val
                    - decrease * (relative_diff_to_bound * abs(previous_val - lb))
                    + increase * (relative_diff_to_bound * abs(previous_val - ub))
                )
                val_e = (
                    val_s
                    + range_decrease * (range_diff * abs(val_s - lb))
                    + range_increase * (range_diff * abs(val_s - ub))
                )
                previous_val = val_e
                vals_s.append(val_s)
                vals_e.append(val_e)
        working[self.target_start_variable] = vals_s
        working[self.target_end_variable] = vals_e

        # optional rounding step
        if self.should_round_output:
            working[self.target_start_variable] = working[
                self.target_start_variable
            ].astype(int)
            working[self.target_end_variable] = working[
                self.target_end_variable
            ].astype(int)

        working = working.sort_values(by=["original_order"])

        # remove tmp variables
        columns_to_remove = [
            "increase",
            "decrease",
            self.new_difference_variable_name,
            self.new_first_variable_name,
            self.new_range_variable,
            "range_increase",
            "range_decrease",
            "original_order",
        ]
        working = working.drop(columns=columns_to_remove)

        # order columns
        common_cols = [c for c in source.columns if c in working.columns]
        other_cols = [c for c in working.columns if c not in source.columns]
        cols_order = common_cols + other_cols
        working = working[cols_order]

        return working



from typing import List

import pandas as pd

CATEGORICAL_DTYPES = ["object", "category", "boolean"]


def get_continuous_under_threshold(
    df: pd.DataFrame, *, threshold: int = 10
) -> List[str]:
    """Get continuous variable names with number of unique values under a threshold."""
    columns_to_check = df.select_dtypes(exclude=CATEGORICAL_DTYPES).columns
    return list(filter(lambda col: df[col].nunique() <= threshold, columns_to_check))



import numpy as np
import pandas as pd
import pandas.testing as pd_testing
import pytest

from avatars.processors import ProportionProcessor

ORIGINAL_DF = pd.DataFrame(
    {
        "variable_1": [100, 150, 120, 100],
        "variable_2": [10, 30, 30, np.nan],  # to test v2+v3+v4 = v1
        "variable_3": [30, 60, 30, np.nan],
        "variable_4": [60, 60, 60, np.nan],
        "variable_5": [90, 90, 90, np.nan],
    }
)
TRANSFORMED_DF = pd.DataFrame(
    {
        "variable_1": [100, 150, 120, 100],
        "variable_2": [0.1, 0.2, 0.25, np.nan],  # to test v2+v3+v4 = v1
        "variable_3": [0.3, 0.4, 0.25, np.nan],
        "variable_4": [0.6, 0.4, 0.5, np.nan],
        "variable_5": [0.9, 0.6, 0.75, np.nan],
    }
)


def test_preprocess_proportional_variables() -> None:
    # Check that three variables can be transformed into proportion of a fourth one.
    processor = ProportionProcessor(
        variable_names=["variable_2", "variable_3", "variable_4"],
        reference="variable_1",
        sum_to_one=True,
    )
    df = processor.preprocess(df=ORIGINAL_DF)
    pd_testing.assert_frame_equal(
        TRANSFORMED_DF[["variable_1", "variable_2", "variable_3", "variable_4"]],
        df[["variable_1", "variable_2", "variable_3", "variable_4"]],
    )


def test_postprocess_proportional_variables() -> None:
    # Check that three variables expressed as proportion can be re-expressed in their
    # original form.
    processor = ProportionProcessor(
        variable_names=["variable_2", "variable_3", "variable_4"],
        reference="variable_1",
        sum_to_one=True,
    )
    df = processor.postprocess(source=None, dest=TRANSFORMED_DF)
    pd_testing.assert_frame_equal(
        ORIGINAL_DF[["variable_1", "variable_2", "variable_3", "variable_4"]],
        df[["variable_1", "variable_2", "variable_3", "variable_4"]],
        check_dtype=False,
    )


def test_preprocess_proportional_single_variable() -> None:
    # Check that 1 variable can be transformed into proportion of another.
    processor = ProportionProcessor(
        variable_names=["variable_5"], reference="variable_1", sum_to_one=False
    )
    df = processor.preprocess(df=ORIGINAL_DF)
    pd_testing.assert_frame_equal(
        TRANSFORMED_DF[["variable_1", "variable_5"]],
        df[["variable_1", "variable_5"]],
    )


def test_postprocess_proportional_single_variable() -> None:
    # Check that 1 variable can be transformed into proportion of another.
    processor = ProportionProcessor(
        variable_names=["variable_5"], reference="variable_1", sum_to_one=False
    )
    df = processor.postprocess(source=None, dest=TRANSFORMED_DF)
    pd_testing.assert_frame_equal(
        ORIGINAL_DF[["variable_1", "variable_5"]],
        df[["variable_1", "variable_5"]],
        check_dtype=False,
    )


def test_preprocess_wrong_reference() -> None:
    processor = ProportionProcessor(
        variable_names=["variable_2"],
        reference="wrong_variable",
        sum_to_one=True,
    )
    with pytest.raises(
        ValueError,
        match="('variable_name', "
        "'variable wrong_variable cannot be found in the dataframe variables')",
    ):
        processor.preprocess(ORIGINAL_DF)


def test_postprocess_wrong_reference() -> None:
    processor = ProportionProcessor(
        variable_names=["variable_2"],
        reference="wrong_variable",
        sum_to_one=True,
    )
    with pytest.raises(
        ValueError,
        match="('variable_name', "
        "'variable wrong_variable cannot be found in the dataframe variables')",
    ):
        processor.postprocess(ORIGINAL_DF, ORIGINAL_DF)


def test_preprocess_reference_variable_zero() -> None:
    # Verify result contains proportions when reference is zero.
    df = pd.DataFrame(
        {
            "variable_1": [100, 0],
            "variable_2": [10, 10],
            "variable_3": [90, 30],
        }
    )
    processor = ProportionProcessor(
        variable_names=["variable_2", "variable_3"],
        reference="variable_1",
        sum_to_one=True,
    )
    preprocessed = processor.preprocess(df=df)
    expected = pd.DataFrame(
        {
            "variable_1": [100, 0],
            "variable_2": [0.1, 0.25],
            "variable_3": [0.9, 0.75],
        }
    )
    pd_testing.assert_frame_equal(preprocessed, expected)


def test_preprocess_target_variables_sum_zero() -> None:
    # Verify target variables that equal zero still equal zero after transformation
    df = pd.DataFrame(
        {
            "variable_1": [100, 100],
            "variable_2": [10, 0],
            "variable_3": [90, 0],
        }
    )
    processor = ProportionProcessor(
        variable_names=["variable_2", "variable_3"],
        reference="variable_1",
        sum_to_one=True,
    )
    preprocessed = processor.preprocess(df=df)
    expected = pd.DataFrame(
        {
            "variable_1": [100, 100],
            "variable_2": [0.1, 0],
            "variable_3": [0.9, 0],
        }
    )
    pd_testing.assert_frame_equal(preprocessed, expected)


def test_postprocess_saferounding() -> None:
    "Verify that the postprocessed variables sum up to reference variable."
    df = pd.DataFrame(
        {
            "variable_1": [101, 101],
            "variable_2": [34, 20],
            "variable_3": [66, 80],
        }
    )
    preprocessed = pd.DataFrame(
        {
            "variable_1": [101, 101],
            "variable_2": [0.34, 0.20],
            "variable_3": [0.66, 0.80],
        }
    )
    processor = ProportionProcessor(
        variable_names=["variable_2", "variable_3"],
        reference="variable_1",
        sum_to_one=True,
        decimal_count=0,
    )
    postprocessed = processor.postprocess(df, preprocessed)
    sum_check = postprocessed["variable_2"] + postprocessed["variable_3"]
    pd_testing.assert_series_equal(
        sum_check, postprocessed["variable_1"], check_names=False, check_dtype=False
    )


def test_decimals() -> None:
    "Verify that the postprocessed variables have the specified number of decimals."
    df = pd.DataFrame(
        {
            "variable_1": [101, 101],
            "variable_2": [34, 20],
            "variable_3": [66, 80],
        }
    )
    preprocessed = pd.DataFrame(
        {
            "variable_1": [101, 101],
            "variable_2": [0.34, 0.20],
            "variable_3": [0.66, 0.80],
        }
    )
    processor = ProportionProcessor(
        variable_names=["variable_2", "variable_3"],
        reference="variable_1",
        sum_to_one=True,
        decimal_count=2,
    )
    postprocessed = processor.postprocess(df, preprocessed)
    expected = pd.DataFrame(
        {
            "variable_1": [101, 101],
            "variable_2": [34.34, 20.20],
            "variable_3": [66.66, 80.80],
        }
    )
    pd_testing.assert_frame_equal(postprocessed, expected)
